/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GRABBERWIDGET_H
#define GRABBERWIDGET_H

#include "GLUtilityMethods.h"
#include <QGLViewer/qglviewer.h>
#include "UserStroke.h"
#include <QPoint>

/**
 * @brief The SelectionWidget class
 * widget used to capture the user strokes and perform the selections
 */
class SelectionWidget : public QObject, public qglviewer::MouseGrabber
{
    Q_OBJECT
public :
    /**
     * @brief The STATE enum the widget can be either active or inactive
     */
    enum STATE {INACTIVE = 0, ACTIVE = 1};

    /**
     * @brief The MODE enum the widget can be used to draw, erase, delete or refine a selection
     */
    enum MODE {DRAW = 0, ERASE = 1, DEL = 2, REFINE = 3};
protected:

    STATE _state; /**< state of the widget (active or not)*/
    MODE _mode; /**< widget mode */
    UserStroke _stroke; /**< user stroke */
    UserStroke _refineStroke; /**< user stroke to refine the selection */
    float previous_dist; /**< */
    Vec3Df previous_p; /**< */

    bool _mouse_released;  /**< set to true when the user releases the mouse */
    bool selectionStarted; /**< set to true when the selection is started */
    Vec3Df _Uplan; /**<  use to draw the widget in the screen plane*/
    Vec3Df _Vplan; /**<  use to draw the widget in the screen plane*/

    QPoint lastPointAdded; /**< last point added (in screen space) added */

    Vec3Df center; /**< widget center */
    float radius; /**< widget radius */

public:

    SelectionWidget()
    {
        _state = INACTIVE;
        _mouse_released = true;
        selectionStarted = false;
        _mode = DRAW;
        center = Vec3Df(0.f, 0.f, 0.f);
        radius = 0.02;
    }

    ~SelectionWidget(){}

    /**
     * @brief setState set the widget state to s
     * @param s
     */
    void setState(int s)
    {
        Trace(4, " SelectionWidget::setState "<< _mouse_released);
        _state = (STATE) s;
        if (_state == ACTIVE)
            _stroke.deleteAllPoints();
        else
            _mouse_released = true;
    }

    /**
     * @brief getState
     * @return the current state of the widget
     */
    int getState() { return (int) _state; }

    /**
     * @brief setMode set the mode of the widget to m
     * @param m
     */
    void setMode (int m){
        _mode = (MODE) m;
    }

    /**
     * @brief getMode
     * @return return the current mode of the widget
     */
    int getMode(){return (int) _mode;}

    /**
     * @brief clear the widget (erase previous strokes)
     */
    void clear()
    {
        _stroke.deleteAllPoints();
        _refineStroke.deleteAllPoints();
        lastPointAdded = QPoint(-100, -100);
    }

    /**
     * @brief setColor set the widget color to color
     * @param color
     */
    void setColor(QColor color)
    {
        _stroke.setColor(color);
    }

    /**
     * @brief setRadius set the radius widget to r
     * @param r
     */
    void setRadius(float r){radius = r;}

    /**
     * @brief draw draw the widget and the stroke
     * @param color color of the stroke
     * @param drawStroke true if we have to draw the stroke (part mode)
     */
    void draw(std::vector<float> & color, bool drawStroke)
    {
        if (_state == ACTIVE)
        {
            // draw the stroke
            if ( drawStroke )
                _stroke.draw(color);
            glColor4f(1.f, 0.f, 0.5f, 1.f);
            float teta;
            Vec3Df p;
            glDisable(GL_DEPTH_TEST);

            // draw the widget (here a simple circle)
            glLineWidth(2.f);
            glBegin(GL_LINE_STRIP);
            for(int i=0; i<360; i+=5)
            {
                teta = (float)(i) * M_PI/ float(180);
                p = center + radius*cosf(teta)*_Uplan + radius*sinf(teta)*_Vplan;
                glVertex3f(p[0],p[1],p[2]);
            }
            glEnd();
            glEnable(GL_DEPTH_TEST);
        }
    }

    /**
     * @brief setRepUplan set _Uplan to x
     * @param x
     */
    void setRepUplan(Vec3Df x){_Uplan = x; _Uplan.normalize();}

    /**
     * @brief setRepVplan set _Vplan to y
     * @param y
     */
    void setRepVplan(Vec3Df y){_Vplan = y; _Vplan.normalize();}

    /**
     * @brief getStroke
     * @return  return the user stroke
     */
    UserStroke &  getStroke(){ return _stroke;}

    /**
     * @brief getRefineStroke
     * @return return the refine stroke
     */
    UserStroke &  getRefineStroke(){ return _refineStroke;}

    /**
     * @brief checkIfGrabsMouse check if the widget can grabs the mouse
     * if the widget is active and if the mouse is on the object
     * then the widget can grabs the mouse
     * @param x x value of the mouse position
     * @param y y value of the mouse position
     * @param cam camera of the viewer
     */
    void checkIfGrabsMouse(int x, int y, const qglviewer::Camera* const cam)
    {
        switch(_state)
        {
        case INACTIVE:
            setGrabsMouse(false); // widget inactive
            break;
        case ACTIVE:
            Trace(4, " SelectionWidget::checkIfGrabsMouse "<< _mouse_released);
            if (!_mouse_released)
            {
                setGrabsMouse(true);
                return;
            }
            bool found;
            cam->pointUnderPixel(QPoint(x,y), found);

            if (found)
            {
                setGrabsMouse(true);
                return;
            }
            setGrabsMouse(false);
            return;
        }
    }

    /**
     * @brief addPointToStroke add a point to the stroke
     * @param event contains the mouse information
     * @param cam contains the camera information
     * @return true if a point is added
     */
    bool addPointToStroke (QMouseEvent* const event, qglviewer::Camera* const cam);

    /**
     * @brief addPointToRefineStroke
     * @param event contains the mouse information
     * @param cam  contains the camera information
     * @return true if a point is added
     */
    bool addPointToRefineStroke (QMouseEvent* const event, qglviewer::Camera* const cam);

    /**
     * @brief erasePointsToStroke
     * @param event contains the mouse information
     * @param cam  contains the camera information
     * @return true if a point is added
     */
    bool erasePointsToStroke (QMouseEvent* const event, qglviewer::Camera* const cam);

    /**
     * @brief deleteStrokeUnderPoint
     * @param event contains the mouse information
     * @param cam  contains the camera information
     * @return true if a point is added
     */
    bool deleteStrokeUnderPoint(QMouseEvent* const event, qglviewer::Camera* const cam);

    /**
     * @brief mousePressEvent
     * @param event contains the mouse information
     * @param cam  contains the camera information
     */
    void mousePressEvent(QMouseEvent* const event, qglviewer::Camera* const cam );

    /**
     * @brief mouseMoveEvent
     * @param event contains the mouse information
     * @param cam  contains the camera information
     */
    void mouseMoveEvent(QMouseEvent* const event, qglviewer::Camera* const cam );

    /**
     * @brief mouseReleaseEvent
     * @param event contains the mouse information
     * @param cam  contains the camera information
     */
    void mouseReleaseEvent(QMouseEvent* const event, qglviewer::Camera* const cam );

    /**
     * @brief wheelEvent
     * @param event contains the mouse information
     * @param camera  contains the camera information
     */
    void wheelEvent(QWheelEvent *const event, qglviewer::Camera *const camera);

signals:
    void pressed();
    void moved();
    void released();
    void changeLaplacian();
    void showMSHist(Vec3Df v);
    void eraseMode();
};

#endif // GRABBERWIDGET_H
