/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MESHLAPLACIAN_H
#define MESHLAPLACIAN_H

#include "Mesh.h"

/**
 * @brief The MeshLaplacian class
 * virtual class to compute a mesh laplacian
 */
class MeshLaplacian : public Mesh
{
public:

    MeshLaplacian():Mesh(){}
    virtual ~MeshLaplacian (){}

    /**
     * @brief init initialize from a mesh
     * @param m
     */
    void init(const Mesh &m){
        vertices = m.getVertices();
        triangles = m.getTriangles();
        oneRing = m.getOneRing();
    }

    /**
     * @brief computeWeightAndAreaMaps
     * virtual function to compute the weights (edgesWeightMap) and area (areaMap)
     */
    virtual void computeWeightAndAreaMaps() = 0;

    /**
     * @brief computeLaplacian
     * virtual function to compute the mesh laplacian
     */
    virtual void computeLaplacian() = 0;

    /**
     * @brief getLaplacian
     * @return the mesh laplacian
     */
    const std::vector<std::vector<std::pair<unsigned int, float> > > & getLaplacian(){
        return laplacian;
    }

    /**
     * @brief getWeightMap
     * @return the weight map
     */
    EdgeWeights & getWeightMap(){return edgesWeightMap;}

    /**
     * @brief getAreaMap
     * @return the area map
     */
    EdgeWeights & getAreaMap(){return areaMap;}

protected:
    std::vector< std::vector< std::pair<unsigned int, float> > > laplacian;
    /**< mesh laplacian */
    EdgeWeights edgesWeightMap; /**< edges weights */
    EdgeWeights areaMap; /**< area map*/
};

#endif // MESHLAPLACIAN_H
