/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <iostream>
#include <Vec3D.h>
#include <vector>

/**
 * @brief The Triangle class
 */
class Triangle {
public:
    /**
     * @brief Triangle  build a triangle from 3 vertices
     * @param v0
     * @param v1
     * @param v2
     */
    inline Triangle(unsigned int v0, unsigned int v1, unsigned int v2)
    { init (v0, v1, v2); }

    /**
     * @brief Triangle built a triangle from a list of indices
     * @param vp
     */
    inline Triangle (unsigned int * vp) { init (vp[0], vp[1], vp[2]); }

    /**
     * @brief Triangle build a triangle from another one
     * @param it
     */
    inline Triangle (const Triangle & it) {
        init (it.v[0], it.v[1], it.v[2]);
        norm = it.norm;
        smoothNorm = it.smoothNorm;
    }

    inline virtual ~Triangle () {}

    /**
     * @brief operator =
     * @param it triangle to copy
     * @return this = it.
     */
    inline Triangle & operator= (const Triangle & it) {
        init (it.v[0], it.v[1], it.v[2]);
        norm = it.norm;
        smoothNorm = it.smoothNorm;
        return (*this);
    }

    /**
     * @brief operator ==
     * @param t
     * @return true if 2 triangles are equals
     */
    inline bool operator== (const Triangle & t) const {
        return (v[0] == t.v[0] && v[1] == t.v[1] && v[2] == t.v[2]);
    }
    /**
     * @brief getVertex
     * @param i
     * @return return vertex i
     */
    inline unsigned int getVertex (unsigned int i) const { return v[i]; }

    /**
     * @brief setVertex
     * @param i
     * @param vertex
     */
    inline void setVertex (unsigned int i, unsigned int vertex) { v[i] = vertex; }

    /**
     * @brief contains
     * @param vertex
     * @return true if vertex is a one of the triangle vertices
     */
    inline bool contains (unsigned int vertex) const {
        return (v[0] == vertex || v[1] == vertex || v[2] == vertex);
    }

    /**
     * @brief isNeighbor
     * @param t
     * @return true if the 2 triangles share exactly 2 vertices
     */
    inline bool isNeighbor(Triangle t){
        for (unsigned int j = 0; j < 3; j ++)
        {
            if (this->contains(t.v[j]) && this->contains(t.v[(j+1)%3]) && !this->contains(t.v[(j+2)%3]))
                return true;
        }
        return false;
    }

    /**
     * @brief getNorm
     * @return return the triangle normal
     */
    inline const Vec3Df & getNorm () const { return norm; }

    /**
     * @brief getSmoothNorm
     * @return return the smoothed triangle normal
     */
    inline const Vec3Df & getSmoothNorm () const { return smoothNorm; }

    /**
     * @brief setNorm set the triangle Normal to _norm
     * @param _norm
     */
    inline void setNorm (const Vec3Df & _norm){ norm = _norm;}

    /**
     * @brief setSmoothNorm set the smoothed triangle normal to _norm
     * @param _norm
     */
    inline void setSmoothNorm (const Vec3Df & _norm){ smoothNorm = _norm;}

protected:
    /**
     * @brief init
     * @param v0
     * @param v1
     * @param v2
     */
    inline void init (unsigned int v0, unsigned int v1, unsigned int v2) {
        v[0] = v0; v[1] = v1; v[2] = v2;
    }

private:
    unsigned int v[3]; /**< indice of the triangle vertices */
    Vec3Df norm; /**< triangle norm */
    Vec3Df smoothNorm; /**< triangle smoothed norm */
    //std::vector<int> voisins;
};

extern std::ostream & operator<< (std::ostream & output, const Triangle & t);

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
