/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CHOLMODLSSTRUCT_H
#define CHOLMODLSSTRUCT_H


#include "cholmod.h"
#include <QTime>

/**
 * @brief The CholmodLSStruct class
 *  solve AtAx = Atb
 *  use the fast update/downdate technique
 */
class CholmodLSStruct
{

    cholmod_common _c; /**< contains the common parameters, statistics, and workspace needed by cholmod*/
    cholmod_triplet *_triplet; /**< sparse matrix in triplet form */

    int *_rowPtrA; /**< pointer toward the row indices of A */
    int *_colPtrA; /**< pointer toward the column indices of A */
    double *_valuePtrA; /**< pointer toward the nnz values of A */

    cholmod_sparse *_At; /**< A transpose in sparse format*/
    cholmod_factor *_L; /**< contains the factorization of the system*/
    cholmod_sparse* _A; /**< _A matrix */
    cholmod_dense* _Atb; /**< _Atb matrix */
    cholmod_sparse* _AtA; /**< _AtA matrix */

    cholmod_sparse* _Cm; /**< _Cm constraint matrix for the downdate/update */
    cholmod_sparse* _Ct; /**< _Ct constraint matrix for the downdate/update */
    int *_rowPtrC; /**< pointer toward the row indices of Cm */
    int *_colPtrC; /**< pointer toward the column indices of Cm */
    double* _valuePtrC; /**< pointer toward the values of Cm */
    int  _nb_non_zeros_in_C , _Crows , _Ccols;
    cholmod_triplet *_tripletC;
    cholmod_dense* _deltaB; /**< updated values of b*/
    cholmod_dense* _CtdeltaB; /**<  */
    double *_valuePtrDeltaB;

    cholmod_dense *_b;
    double *_valuePtrB;

    cholmod_dense* _x;
    double * _x_data;

    int _nb_non_zeros_in_A , _Arows , _Acols , _bcols;
    // b has _Arows rows , and _bcols columns;
    bool allocated , factorized, allocatedUp;

private:
    inline void resetDimensions()
    {
        setDimensions( -1 , -1 , -1 );
        setCDimensions(-1, -1);
    }
    inline void resetNNZinA()
    {
        _nb_non_zeros_in_A = -1;
    }

    inline void resetNNZinC()
    {
        _nb_non_zeros_in_C = -1;
    }

public:
    CholmodLSStruct() :  _nb_non_zeros_in_C(-1),   _Crows(-1),_Ccols(-1),
        _nb_non_zeros_in_A(-1) , _Arows(-1) , _Acols(-1) , _bcols(-1) ,
        allocated(false) , factorized(false),
        allocatedUp(false) {}


    /**
     * @brief start start cholmod with all the common parameters and workspace..
     * must be the first call to cholmod
     */
    void start()
    {
        cholmod_start(&_c);
    }

    void finish()
    {
    }

    /**
     * @brief setDimensions
     * @param rA number of rows in A
     * @param cA number of columns in A
     * @param cB number of columns in B
     */
    void setDimensions( int rA , int cA , int cB )
    {
        _Arows = rA;
        _Acols = cA;
        _bcols = cB;
    }

    /**
     * @brief setCDimensions
     * @param rC number of rows
     * @param cC number of columns
     */
    void setCDimensions(int rC, int cC)
    {
        _Crows = rC;
        _Ccols = cC;
    }

    /**
     * @brief specifyNonZeroInA
     * @param nnz number of non zero elements in A
     */
    void specifyNonZeroInA( int nnz )
    {
        _nb_non_zeros_in_A = nnz;
    }

    /**
     * @brief specifyNonZeroInC
     * @param nnz number of non zero elements in C
     */
    void specifyNonZeroInC(int nnz)
    {
        _nb_non_zeros_in_C = nnz;
    }

    /**
     * @brief addValueInA add the nnz value val in A
     * @param r row indice
     * @param c column indice
     * @param v value
     */
    void addValueInA( int r , int c , double v )
    {
        const int i = _triplet->nnz;
        _rowPtrA[i] = r;
        _colPtrA[i] = c;
        _valuePtrA[i] = v;
        _triplet->nnz++;
    }

    /**
     * @brief addValueInC add the nnz value val in C
     * @param r row indice
     * @param c column indice
     * @param v value
     */
    void addValueInC( int r , int c , double v )
    {
        const int i = _tripletC->nnz;
        _rowPtrC[i] = r;
        _colPtrC[i] = c;
        _valuePtrC[i] = v;
        _tripletC->nnz++;
    }

    /**
     * @brief setValueInB
     * @param r row index
     * @param c column index
     * @param v value
     */
    void setValueInB( int r , int c , double v )
    {
        // Cholmod specs : Entry in row i and column j is located in x [i+j*d]
        _valuePtrB[ r + c * _Arows ] = v;
    }

    /**
     * @brief setValueInDeltaB (delta b is used as the new b when the system is updated)
     * @param r row index
     * @param c column index
     * @param v value
     */
    void setValueInDeltaB(int r, int c, double v){
        _valuePtrDeltaB[r + c * _Crows] = v;
    }

    /**
     * @brief allocate allocate all the space needed by cholmod
     * to perform the factorization
     */
    void allocate()
    {
        if( allocated ) return;

        _triplet = cholmod_allocate_triplet(
                    _Arows ,
                    _Acols ,
                    _nb_non_zeros_in_A ,
                    0, CHOLMOD_REAL, &_c);

        _rowPtrA = (int*)_triplet->i;
        _colPtrA = (int*)_triplet->j;
        _valuePtrA = (double*)_triplet->x;

        _b = cholmod_zeros(_Arows, _bcols, CHOLMOD_REAL, &_c);
        _valuePtrB = (double*)_b->x;
        _Atb = cholmod_allocate_dense(_Acols, _bcols, _bcols * _Acols, CHOLMOD_REAL, &_c);

        allocated = true;
        factorized = false;
    }

    /**
     * @brief free free all the space that was allocated before
     */
    void free()
    {
        if( !allocated ) return;

        cholmod_free_triplet(&_triplet,&_c);

        if( factorized )
            cholmod_free_sparse(&_A,&_c);

        cholmod_free_sparse(&_At,&_c);
        cholmod_free_sparse(&_AtA, &_c);
        cholmod_free_factor(&_L,&_c);

        cholmod_free_dense(&_b,&_c);

        allocated = false;
        factorized = false;

        if (allocatedUp)
            cholmod_free_triplet(&_tripletC,&_c);

        cholmod_free_dense(&_Atb,&_c);
        cholmod_free_dense(&_x,&_c);

        resetDimensions();
        resetNNZinA();
        resetNNZinC();
    }

    /**
     * @brief freeSolution
     */
    void freeSolution()
    {
        cholmod_free_dense(&_x, &_c);
        _x_data = NULL;
    }

    /**
     * @brief allocateUpdate allocate space needed by cholmod to perform
     *an update of the factorization
     */
    void allocateUpdate(){
        if (allocatedUp)
            return;

        _tripletC = cholmod_allocate_triplet(
                    _Crows,
                    _Ccols,
                    _nb_non_zeros_in_C,
                    0, CHOLMOD_REAL, &_c);
        _rowPtrC = (int*)_tripletC->i;
        _colPtrC = (int*)_tripletC->j;
        _valuePtrC = (double*)_tripletC->x;

        _deltaB = cholmod_zeros(_Crows, 1, CHOLMOD_REAL, &_c);
        _CtdeltaB = cholmod_allocate_dense(_Ccols, _bcols, _bcols * _Ccols,
                                           CHOLMOD_REAL, &_c);
        _valuePtrDeltaB = (double*)_deltaB->x;

        allocatedUp = true;
    }

    /**
     * @brief freeUpdate free the space that was needed by cholmod to perform
     *an update of the factorization
     */
    void freeUpdate(){
        if (!allocatedUp)
            return;
        Trace(4,  " free update begin \n ");
        cholmod_free_sparse(&_Ct,&_c);
        cholmod_free_sparse(&_Cm, &_c);
        cholmod_free_triplet(&_tripletC, &_c);
        cholmod_free_dense(&_CtdeltaB, &_c);
        cholmod_free_dense(&_deltaB, &_c);
        resetNNZinC();
        allocatedUp = false;
        Trace(4,  " free update end \n ");
    }

    /**
     * @brief getSolutionValue
     * @param r row index
     * @param c column index
     * @return return the value of the solution for r and c
     */
    double getSolutionValue( int r , int c )
    {
        return _x_data[r + c * _Acols];
    }

    /**
     * @brief factorize the system _AtA
     */
    void factorize()
    {
        // Careful : in this function, there is an allocation, you need to take care of the memory

        // compute the At*A matrix
        _A   = cholmod_triplet_to_sparse(_triplet, _triplet->nnz, &_c);
        _At  = cholmod_transpose(_A, 1, &_c);
        _AtA = cholmod_ssmult(_At, _A, 0, 1, 1, &_c);
        _AtA->stype = 1; // matrix is square and symmetric

        // perform a symbolic factorization
        _L = cholmod_analyze(_AtA, &_c);

        // numeric factorization
        // _AtA matrix to factorize
        // _L resulting factorization
        cholmod_factorize(_AtA, _L, &_c);

        factorized = true;
    }

    /**
     * @brief solve solve AtAx = b
     * @param coutResult true if we want to print the time needed to solve the system
     */
    void solve( bool coutResult = true )
    {
        QTime timer_solve;
        timer_solve.start();

        double alpha[] = {1, 1};
        double beta[] = {0, 0};
        cholmod_sdmult(_At, 0, alpha, beta, _b, _Atb, &_c);
        _x = cholmod_solve(CHOLMOD_A, _L, _Atb, &_c);

        if( coutResult )
            std::cout << "cholmod solve took " << timer_solve.elapsed() << " ms" << std::endl;

        // get result:
        _x_data = (double*)(_x->x);
        // Do not forget to free x each time!
       // print_Atb();
    }

    /**
     * @brief update solve the system with the new constraints
     * updates the LDL factorization so that it becomes LDLt + CCt
     * updates the b vector
     * solve the new system
     */
    void update()
    {
        Trace(4, "cholmod struct UPDATE begin ");

        double alpha[] = {1, 1};
        double beta[] = {0, 0};
        _Cm  = cholmod_triplet_to_sparse(_tripletC, _tripletC->nnz, &_c);
        _Ct = cholmod_transpose(_Cm, 1, &_c);

        cholmod_sparse *_Ctnew = cholmod_submatrix (_Ct, (int*) _L->Perm, _L->n, NULL, -1, true, true, &_c) ;
        cholmod_updown(
                    true,
                    _Ctnew,
                    _L,
                    &_c);

        cholmod_free_sparse(&_Ctnew, &_c);

        cholmod_sdmult(_Ct, 0, alpha, beta, _deltaB, _CtdeltaB, &_c);
        _x = cholmod_solve(CHOLMOD_A, _L, _CtdeltaB, &_c);

        // get result:
        _x_data = (double*)(_x->x);

        Trace(4, " cholmod end of update \n");
    }


    /**
     * @brief downdate
     * downdates the LDL factorization so that it becomes LDLt - CCt
     */
    void downdate()
    {
        Trace(4, "cholmod struct downdate begin ");
        _Cm  = cholmod_triplet_to_sparse(_tripletC, _tripletC->nnz, &_c);
        _Ct = cholmod_transpose(_Cm, 1, &_c);

        cholmod_sparse *_Ctnew = cholmod_submatrix (_Ct, (int*) _L->Perm, _L->n, NULL, -1, true, true, &_c) ;
        cholmod_updown(
                    false,
                    _Ctnew,
                    _L,
                    &_c);

        cholmod_free_sparse(&_Ctnew, &_c);
        Trace(4, " cholmod end of downdate \n");
    }

    /**
     * @brief print_Atb
     */
    void print_Atb()
    {
        std::cout << " _Atb values \n ";
        double * val = (double*)(_Atb->x);
        for (unsigned int i = 0; i < _Atb->nrow; i++)
            if (val[i] != 0.)
                std::cout << " i " << i << " : " << val[i] << "\n";
        std::cout << " --------------------------------- \n ";
    }

    /**
     * @brief print_CtDeltaB
     */
    void print_CtDeltaB()
    {
        std::cout << " _AtDeltaB values \n ";
        double * val = (double*)(_CtdeltaB->x);
        for (unsigned int i = 0; i < _CtdeltaB->nrow; i++)
            if (val[i] != 0.)
                std::cout << " i " << i << " : " << val[i] << "\n";
        std::cout << " --------------------------------- \n ";
    }


    // you should start your structure at the beginning of your program,
    // then allocate once and for all with the correct _rows, _cols and _nb_non_zeros_in_A,
    // then once set the values in A, and factorize it,
    // then set each time the values in b, and solve it, and free the result each time
};

#endif // CHOLMODLSSTRUCT_H
