/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DISTANCE_H
#define DISTANCE_H

#include "Vec3D.h"

/**
 * @brief The DIST_TYPE enum
 */
enum DIST_TYPE{
    DIST_L2 = 0,
    DIST_WENDLAND = 1,
    DIST_L21_W = 2,
    DIST_L21 = 3,
    DIST_LP_2 = 4,
    DIST_LP_4 = 5,
    DIST_LP_16 = 6,
    DIST_LP1 = 7,
    DIST_L_CROSS = 8
};

static const unsigned int numberOfDistancesDefined = 9;

/**
 * @brief wendland function ((1-x)/h)^4 * x/(h+1)
 * @param x
 * @param h
 * @return wendland (x, h)
 */
float wendland(float x, float h)
{
    if(x>h)
        return 0;

    float dist = ((1-x/h)*(1-x/h)*(1-x/h)*(1-x/h)*(4*x/h+1));
    assert(dist >= 0);
    return dist;
}

/**
 * @brief distanceWendland wendland distance between two triangles
 * @param iBar triangle i barycenter
 * @param jBar triangle j barycenter
 * @param iNorm triangle i normal
 * @param jNorm  triangle j normal
 * @param normInfluence
 * @return distance between triangle i and j
 */
float distanceWendland (Vec3Df & iBar, Vec3Df & jBar,
                        Vec3Df & iNorm, Vec3Df & jNorm, float normInfluence)
{
    float d = (iBar-jBar).getLength();
    float n = (iNorm-jNorm).getSquaredLength();
    float dist = d*(1+(1-wendland(n,4))*normInfluence*100);
    assert(dist >= 0);
    return dist;
}

/**
 * @brief distanceL2_1weighted
 * @param iBar
 * @param jBar
 * @param iNorm
 * @param jNorm
 * @param normInfluence
 * @return
 */
float distanceL2_1weighted (Vec3Df & iBar, Vec3Df & jBar,
                            Vec3Df & iNorm, Vec3Df & jNorm, float  normInfluence)
{
    float d = (iBar-jBar).getLength();
    float n = (iNorm-jNorm).getSquaredLength();
    float dist = ( (1-normInfluence)*d +normInfluence*n);
    assert(dist >= 0);
    return dist;
}

/**
 * @brief distanceL2_1w
 * @param iBar
 * @param jBar
 * @param iNorm
 * @param jNorm
 * @param normInfluence
 * @return
 */
float distanceL2_1w (Vec3Df & iBar, Vec3Df & jBar,
                     Vec3Df & iNorm, Vec3Df & jNorm, float normInfluence)
{
    float d = (iBar-jBar).getLength();
    float n = (iNorm-jNorm).getSquaredLength();
    float dist = ( d +normInfluence*n);
    assert(dist >= 0);
    return dist;
}

/**
 * @brief distanceL2_cross
 * @param iBar
 * @param jBar
 * @param iNorm
 * @param jNorm
 * @param normInfluence
 * @return
 */
float distanceL2_cross(Vec3Df & iBar, Vec3Df & jBar,
                       Vec3Df & iNorm, Vec3Df & jNorm, float normInfluence)
{
    float d = (iBar-jBar).getLength();
    float v = Vec3Df::dotProduct(iNorm,(iBar-jBar));
    float n = (1+std::abs(v)/d)*(iNorm-jNorm).getLength();
    float dist = 0.;

    if ( iBar!=jBar )
        dist = ( d +normInfluence*n);

    assert(dist >= 0);
    return dist;
}

/**
 * @brief distanceL2
 * @param iBar
 * @param jBar
 * @return
 */
float distanceL2 (Vec3Df & iBar, Vec3Df & jBar)
{
    return (iBar-jBar).getLength();
}

/**
 * @brief distanceLp
 * @param iBar
 * @param jBar
 * @param eigenVectors
 * @param eigenValues
 * @param p
 * @return
 */
float distanceLp(Vec3Df & iBar, Vec3Df & jBar,
                 Vec3Df* eigenVectors, float* eigenValues,
                 int p)
{

    float max = std::max(eigenValues[0],std::max(eigenValues[1], eigenValues[2]));


    Vec3Df x1_base = Vec3Df(eigenValues[0]/max*(eigenVectors[0])[0]*iBar[0] + eigenValues[1]/max*(eigenVectors[1])[0]*iBar[1] + eigenValues[2]/max*(eigenVectors[2])[0]*iBar[2],
                            eigenValues[0]/max*(eigenVectors[0])[1]*iBar[0] + eigenValues[1]/max*(eigenVectors[1])[1]*iBar[1] + eigenValues[2]/max*(eigenVectors[2])[1]*iBar[2],
                            eigenValues[0]/max*(eigenVectors[0])[2]*iBar[0] + eigenValues[1]/max*(eigenVectors[1])[2]*iBar[1] + eigenValues[2]/max*(eigenVectors[2])[2]*iBar[2]);

    Vec3Df x_base = Vec3Df( eigenValues[0]/max*(eigenVectors[0])[0]*jBar[0] + eigenValues[1]/max*(eigenVectors[1])[0]*jBar[1] + eigenValues[2]/max*(eigenVectors[2])[0]*jBar[2],
                            eigenValues[0]/max*(eigenVectors[0])[1]*jBar[0] + eigenValues[1]/max*(eigenVectors[1])[1]*jBar[1] + eigenValues[2]/max*(eigenVectors[2])[1]*jBar[2],
                            eigenValues[0]/max*(eigenVectors[0])[2]*jBar[0] + eigenValues[1]/max*(eigenVectors[1])[2]*jBar[1] + eigenValues[2]/max*(eigenVectors[2])[2]*jBar[2]);

    float result = 0.f;

    for (unsigned int k = 0; k < 3; k++)
        result += std::pow((x_base[k]-x1_base[k]),p);
    return (std::pow(result,1/((float) p)));
}

/**
 * @brief distance
 * @param iBar
 * @param jBar
 * @param iNorm
 * @param jNorm
 * @param eigenVectors
 * @param eigenValues
 * @param normInfluence
 * @param weight_dist
 * @return
 */
float distance(Vec3Df & iBar, Vec3Df & jBar, Vec3Df & iNorm, Vec3Df & jNorm,
               Vec3Df* eigenVectors, float* eigenValues, float normInfluence, float*  weight_dist)
{
    float dist = 0.;

    if (weight_dist[DIST_L2]!=0)
        dist += weight_dist[DIST_L2]*distanceL2(iBar,jBar);
    if (weight_dist[DIST_WENDLAND]!=0)
        dist += weight_dist[DIST_WENDLAND]*distanceWendland(iBar, jBar, iNorm, jNorm, normInfluence);
    if (weight_dist[DIST_L21_W]!=0)
        dist += weight_dist[DIST_L21_W]*distanceL2_1weighted(iBar, jBar, iNorm, jNorm, normInfluence);
    if (weight_dist[DIST_L21]!=0)
        dist += weight_dist[DIST_L21]*distanceL2_1w(iBar, jBar, iNorm, jNorm, normInfluence);
    if (weight_dist[DIST_LP_2]!=0)
        dist += weight_dist[DIST_LP_2]*distanceLp(iBar, jBar, eigenVectors,eigenValues, 2);
    if (weight_dist[DIST_LP_4]!=0)
        dist += weight_dist[DIST_LP_4]*distanceLp(iBar, jBar, eigenVectors,eigenValues, 4);
    if (weight_dist[DIST_LP_16]!=0)
        dist += weight_dist[DIST_LP_16]*distanceLp(iBar, jBar, eigenVectors,eigenValues, 16);
    if (weight_dist[DIST_LP1]!=0)
        dist += weight_dist[DIST_LP1]*distanceLp(iNorm, jNorm, eigenVectors,eigenValues, 16);
    if (weight_dist[DIST_L_CROSS]!=0)
        dist += weight_dist[DIST_L_CROSS]*distanceL2_cross(iBar, jBar, iNorm, jNorm, normInfluence);
    return dist;
}


#endif // DISTANCE_H
