/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Mesh.h"
#include <algorithm>
#include <queue>

using namespace std;

void Mesh::clear () {
    Trace(4, "Mesh::clear ");
    clearTopology ();
    clearGeometry ();
}

void Mesh::clearGeometry () {
    Trace(4, "Mesh::clearGeometry ");
    vertices.clear ();
}

void Mesh::clearTopology () {
    Trace(4, "Mesh::clearTopology ");
    triangles.clear ();
}

void Mesh::update()
{
    Trace(4, "Mesh::update ");
    computeBoundingBox();
    Trace(4,  "Mesh : " << vertices.size() << " vertices, " << triangles.size() << " triangles ");
}

void Mesh::computeMeshProperties()
{
    Trace(2, "Mesh::computeMeshProperties ");
    computeBoundingBox();
    Trace(1, "#vertices " << vertices.size() << " #triangles: " <<triangles.size());

    // one Rings stuffs
    computeOneRing();
    computeOneRingTr();
    findFacesNeighbors();

    // normals stuffs
    computeTriangleNormals();
    recomputeSmoothVertexNormals(2);
    computeVerticeSmoothNormals();
    smoothTriangleNormals();

    // connected components
    vector<unsigned int> v;
    for (unsigned int i = 0; i < vertices.size(); i++)
        v.push_back(i);
    findConnectedComponents(vertexConnectedComp, connectedCompRepresentatives, v, oneRing);

    findShapeBoundaries(vertexIsOnShapeBoundary);
}

Vec3Df Mesh::getBar(int t)
{
    Vec3Df v1(vertices[triangles[t].getVertex(0)].getPos());
    Vec3Df v2(vertices[triangles[t].getVertex(1)].getPos());
    Vec3Df v3(vertices[triangles[t].getVertex(2)].getPos());
    return Vec3Df((v1+v2+v3)/3);
}

Vec3Df Mesh::getNorm(int t)
{
    return triangles[t].getNorm();
}

void Mesh::computeVerticeSmoothNormals(){
    Trace(2, "Mesh::computeVerticeSmoothNormals ");
    Vec3Df n, n_i;
    for (unsigned int v = 0; v < vertices.size(); v++)
    {
        n = vertices[v].getNormal();
        for (unsigned int i = 0; i < oneRing[v].size(); i++)
        {
            n_i = vertices[oneRing[v][i]].getNormal();
            float d = (vertices[v].getNormal()-n_i).getSquaredLength();
            n += (std::exp(d/0.245))*n_i;
        }
        n.normalize();
        vertices[v].setSmoothNormal(n);
    }
}

void Mesh::recomputeSmoothVertexNormals (unsigned int normWeight) {
    Trace(2, "Mesh::recomputeSmoothVertexNormals ");
    for (vector<Vertex>::iterator it = vertices.begin ();
         it != vertices.end ();
         it++)
        it->setNormal (Vec3Df (0.0, 0.0, 0.0));

    vector<Triangle>::iterator it = triangles.begin ();

    for ( ; it != triangles.end (); it++)
        for (unsigned int  j = 0; j < 3; j++) {
            Vertex & vj = vertices[it->getVertex (j)];
            float w = 1.0; // uniform weights
            Vec3Df e0 = vertices[it->getVertex ((j+1)%3)].getPos ()
                - vj.getPos ();
            Vec3Df e1 = vertices[it->getVertex ((j+2)%3)].getPos ()
                - vj.getPos ();
            if (normWeight == 1) { // area weight
                w = Vec3Df::crossProduct (e0, e1).getLength () / 2.0;
            } else if (normWeight == 2) { // angle weight
                e0.normalize ();
                e1.normalize ();
                w = (2.0 - (Vec3Df::dotProduct (e0, e1) + 1.0)) / 2.0;
            }
            if (w <= 0.0)
                continue;
            vj.setNormal (vj.getNormal () + (*it).getNorm() * w);
        }
    Vertex::normalizeNormals (vertices);
}

void Mesh::computeTriangleNormals () {
    Trace(2, "Mesh::computeTriangleNormals ");
    for (vector<Triangle>::iterator it = triangles.begin (); it != triangles.end (); it++)
    {
        Vec3Df e01 (vertices[it->getVertex (1)].getPos ()
                    - vertices[it->getVertex (0)].getPos ());
        Vec3Df e02 (vertices[it->getVertex (2)].getPos ()
                    - vertices[it->getVertex (0)].getPos ());
        Vec3Df n (Vec3Df::crossProduct (e01, e02));
        n.normalize ();
        (*it).setNorm(n);
    }
}

void Mesh::smoothTriangleNormals() {
    Trace(2, "Mesh::smoothTriangleNormals ");
    vector<unsigned int> neighbors;
    Vec3Df n;
    for( unsigned int t = 0 ; t < triangles.size(); ++t )
    {
        neighbors = faceNeighbors[t];
        n = triangles[t].getNorm();
        for (unsigned int i = 0; i < neighbors.size(); i++)
        {
            float d = (triangles[i].getNorm()-triangles[neighbors[i]].getNorm()).getSquaredLength();
            n += (std::exp(-d/0.245))*triangles[neighbors[i]].getNorm();
        }
        n.normalize();
        triangles[t].setSmoothNorm(n);
    }
}

void Mesh::computeBoundingBox(){
    Trace(2, "Mesh::computeBoundingBox ");
    BBMin = Vec3Df(FLT_MAX, FLT_MAX, FLT_MAX);
    BBMax = Vec3Df(-FLT_MAX, -FLT_MAX, -FLT_MAX);

    for( unsigned int i = 0; i < vertices.size(); i++){
        Vec3Df point = vertices[i].getPos();

        for( int v = 0 ; v < 3 ; v++ ){
            BBMin[v] = std::min(BBMin[v], point[v]);
            BBMax[v] = std::max(BBMax[v], point[v]);
        }
    }

}

void Mesh::computeAveragePosAndRadius ( Vec3Df & center, float & radius)
{
    Trace(2, "Mesh::computeAveragePosAndRadius ");
    center = Vec3Df(0.0, 0.0, 0.0);
    for (unsigned int i = 0; i < vertices.size (); i++)
        center += (vertices[i]).getPos();
    center /= float (vertices.size ());
    radius = 0.0;
    for (unsigned int i = 0; i < vertices.size (); i++) {
        float vDistance = ((vertices[i]).getPos() - center).getLength();
        if (vDistance > radius)
            radius = vDistance;
    }
}

void Mesh::computeAverageEdgeHalfSize()
{
    Trace(2, "Mesh::computeAverageEdgeHalfSize ");
    average_edge_halfsize = 0.;
    for( unsigned int t = 0 ; t < triangles.size(); ++t )
    {
        for( int i = 0 ; i < 3 ; i ++ )
        {
            float elenght = ((vertices[triangles[t].getVertex((i+1)%3)]).getPos() -
                             (vertices[triangles[t].getVertex(i)]).getPos() ).getLength();
            average_edge_halfsize += elenght;
        }
    }
    average_edge_halfsize = average_edge_halfsize/(2.*triangles.size()*3.);
}

void Mesh::findShapeBoundaries(std::vector<bool> &vertexIsOnShapeBoundary ){
    Trace(2, "Mesh::findShapeBoundaries");
    vertexIsOnShapeBoundary.clear();
    for (unsigned int i = 0; i < vertices.size(); i++)
        vertexIsOnShapeBoundary.push_back(false);

    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        if ( faceNeighbors[i].size() > 0 && faceNeighbors[i].size() < 3)
        {
            unsigned int v0 = triangles[i].getVertex(0);
            unsigned int v1 = triangles[i].getVertex(1);
            unsigned int v2 = triangles[i].getVertex(2);
            unsigned int nb0 = 0 , nb1 = 0, nb2 = 0;
            for (unsigned int j = 0; j < faceNeighbors[i].size(); j++)
            {
                Triangle & t = triangles[faceNeighbors[i][j]];
                if (t.contains(v0))
                    nb0++;
                if (t.contains(v1))
                    nb1++;
                if (t.contains(v2))
                    nb2++;
            }
            if (nb0 < 2)
                vertexIsOnShapeBoundary[v0] = true;
            if (nb1 < 2)
                vertexIsOnShapeBoundary[v1] = true;
            if (nb2 < 2)
                vertexIsOnShapeBoundary[v2] = true;
        }
    }
}

void Mesh::computeGaussianCurvature( std::vector<float> & gaussianCurvature){
    Trace(2, "Mesh::computeGaussianCurvature ");
    gaussianCurvature.clear();

    float min = FLT_MAX;
    float max = FLT_MIN;

    Vec3Df e0, e1, pos_i;

    int v0, v1;
    float sum;
    for (unsigned int i = 0; i < vertices.size(); i++)
    {
        sum = 0;
        pos_i = vertices[i].getPos();
        for (unsigned int j = 0; j < oneRingTr[i].size(); j++)
        {
            Triangle & t = triangles[oneRingTr[i][j]];
            v0 = -1;
            v1 = -1;
            for (unsigned int v = 0; v < 3; v++)
            {
                unsigned int vv = t.getVertex(v);
                if (vv != i)
                {
                    if (v0 == -1)
                        v0 = vv;
                    else
                        v1 = vv;
                }
            }
            assert(v0 != -1);
            assert(v1 != -1);
            e0 = vertices[v0].getPos()-pos_i;
            e1 = vertices[v1].getPos()-pos_i;
            e0.normalize(); e1.normalize();
            sum += acos(Vec3Df::dotProduct(e0, e1));
        }
        if (vertexIsOnShapeBoundary[i])
            gaussianCurvature.push_back(M_PI-sum);
         else
            gaussianCurvature.push_back(2*M_PI-sum);
        if (gaussianCurvature[i] > max)
            max = gaussianCurvature[i];
        if (gaussianCurvature[i] < min)
            min = gaussianCurvature[i];
    }

    // normalize the gaussian curvature
    float max_min = max -min;
    for (unsigned int i = 0; i < gaussianCurvature.size(); i++)
        gaussianCurvature[i] = ( gaussianCurvature[i] - min )/ max_min;
}

void Mesh::findConnectedComponents( std::vector<unsigned int> & parts,
                                    std::vector<unsigned int> & representatives,
                                    std::vector<unsigned int> &  elements,
                                    const std::vector<std::vector<unsigned int > > & neighbors)
{
    Trace(2, "Mesh::findConnectedComponents");
    representatives.clear();
    parts.clear();

    // parts: the indice of the connected component of each vertex.
    for (unsigned int i = 0; i < elements.size(); i++)
        parts.push_back(INT_MAX);

    unsigned int cur_rep;
    bool foundV = true;
    while (foundV)
    {
        foundV = false;
        unsigned int i = 0;
        while (i < elements.size() && !foundV)
        {
            if (parts[elements[i]] == INT_MAX)
            {
                foundV = true;
                break;
            }
            i++;
        }
        if (foundV)
        {
            cur_rep = elements[i];
            representatives.push_back(cur_rep);
            parts[cur_rep] = representatives.size()-1;
            std::queue<unsigned int> vToLook;
            vToLook.push(cur_rep);

            unsigned int cur_v;
            while (! vToLook.empty())
            {
                cur_v = vToLook.front();
                vToLook.pop();
                for (unsigned int v = 0; v < neighbors[cur_v].size(); v++)
                {
                    if ( parts[neighbors[cur_v][v]] == INT_MAX)
                    {
                        vToLook.push(neighbors[cur_v][v]);
                        parts[neighbors[cur_v][v]] = representatives.size()-1;
                    }
                }
            }
        }
    }
    Trace (1, "Mesh::number_of_connected_component_found " << representatives.size() );
}

int Mesh::vertexCloseToPoint(Vec3Df point, float radius)
{
    int result = -1;
    float min_dist = FLT_MAX;
    for( unsigned int v = 0 ; v < vertices.size() ; ++v )
    {
        float dist = ((vertices[v]).getPos() - point).getLength();
        if( dist <= radius && dist < min_dist ){
            result = v;
            min_dist = dist;
        }
    }
    return result;
}

int Mesh::triangleCloseToPoint(Vec3Df point, float radius)
{
    int result = -1;
    float min_dist = FLT_MAX;
    Vec3Df c;
    for (unsigned int t=0; t < triangles.size(); t++)
    {
        c = getBar(t);

        float dist = (c-point).getLength();
        if( dist <= radius && dist < min_dist ){
            result = t;
            min_dist = dist;
        }
    }
    return result;
}

bool Mesh::barycentricCoordinates(Vec3Df point, unsigned int  & t, double & d0, double & d1, double &d2)
{
    Vec3Df p0, p1, p2;
    float min_dist = FLT_MAX;
    float dist;
    double cur_d0, cur_d1, cur_d2;
    t = triangles.size() +2;
    for (unsigned int i=0; i < triangles.size(); i++)
    {
        p0 = vertices[triangles[i].getVertex(0)].getPos();
        p1 = vertices[triangles[i].getVertex(1)].getPos();
        p2 = vertices[triangles[i].getVertex(2)].getPos();

        if (point.pointProjectedOnTriangle(p0, p1, p2,cur_d0, cur_d1, cur_d2, dist))
        {
            if (dist < min_dist)
            {
                t = i;
                d0 = cur_d0;
                d1 = cur_d1;
                d2 = cur_d2;
                min_dist = dist;
            }
        }
    }
    if (min_dist != FLT_MAX)
        return true;
    else
        return false;
}


void Mesh::draw(){
    glBegin (GL_TRIANGLES);
    for (unsigned int i=0; i < triangles.size(); i ++){
        const Triangle &t = triangles[i];
        for (unsigned int j = 0; j < 3; j++)
        {
            glNormalVec3Df (vertices[t.getVertex (j)].getNormal());
            glVertexVec3Df (vertices[t.getVertex (j)].getPos ());
        }
    }
    glEnd ();
}

/****************************************************************
 *  PROTECTED METHODS
 **************************************************************** */

float Mesh::triangleArea(unsigned int i)
{
    Vec3Df a = (vertices[triangles[i].getVertex(0)]).getPos();
    Vec3Df b = (vertices[triangles[i].getVertex(1)]).getPos();
    Vec3Df c = (vertices[triangles[i].getVertex(2)]).getPos();

    return ((Vec3Df::crossProduct(b-a,c-a)).getLength()/2.);
}

float Mesh::computeVoronoiArea(unsigned int v)
{
    float area = 0.f;
    Vec3Df bar;
    Vec3Df v1(0.f, 0.f, 0.f);
    Vec3Df v2(0.f, 0.f, 0.f);
    Vec3Df v_pos = vertices[v].getPos();
    unsigned int tmp;
    for (unsigned int i = 0; i < oneRingTr[v].size(); i++)
    {
        Triangle t = triangles[oneRingTr[v][i]];
        for (unsigned int j = 0; j < 3; j++)
        {
            tmp =  t.getVertex(j);
            if (tmp != v)
            {
                if (v1 == Vec3Df(0.f, 0.f, 0.f))
                    v1 = (vertices[tmp].getPos()-vertices[v].getPos())/2;
                else
                    v2 = (vertices[tmp].getPos()-vertices[v].getPos())/2;
            }
        }
        bar = getBar(oneRingTr[v][i]);
        area += (Vec3Df::crossProduct(v1-v_pos,bar-v_pos)).getLength()/2.f +
                (Vec3Df::crossProduct(v2-v_pos,bar-v_pos)).getLength()/2.f;
    }
    return area;
}

void Mesh::computeOneRing()
{
    Trace(2, "Mesh::computeOneRing ");
    oneRing.clear();
    oneRing.resize(vertices.size());
    for( unsigned int i = 0 ; i < triangles.size() ; i ++ )
    {
        const Triangle & triangle = triangles[i];

        for( int v = 0 ; v < 3 ; v++ ){
            unsigned int vj = triangle.getVertex (v);
            for (unsigned int k = 1; k < 3; k++) {
                unsigned int vk = triangle.getVertex ((v+k)%3);

                if (std::find (oneRing[vj].begin(), oneRing[vj].end(), vk) == oneRing[vj].end ())
                    oneRing[vj].push_back (vk);
            }
        }
    }
}

void Mesh::computeOneRingTr()
{
    Trace(2, "Mesh::computeOneRingTr ");
    oneRingTr.clear();
    oneRingTr.resize (vertices.size ());

    for( unsigned int i = 0 ; i < triangles.size() ; i ++ )
    {
        const Triangle & triangle = triangles[i];
        for (unsigned int k = 0; k < 3; k++) {
            unsigned int v = triangle.getVertex(k);
            oneRingTr[v].push_back (i);
        }
    }
}

void Mesh::findFacesNeighbors()
{
    Trace(2, "Mesh::findFacesNeighbors");
    faceNeighbors.clear();

    vector<unsigned int> potential_neighbors;
    vector<unsigned int> neighbors;

    Vertex v;
    unsigned int f;

    // a face A is the neighbor of a face B if and only if they share an edge.
    // ie they have two common vertices.

    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        potential_neighbors.clear();
        neighbors.clear();
        for (unsigned int j = 0; j <3; j++)
        {
            v= vertices[triangles[i].getVertex(j)];
            const vector<unsigned int> & faceV = v.getFaces();
            for (unsigned int k = 0; k < faceV.size(); k++)
            {
                f = faceV[k];
                if (f != i)
                {
                    if ( find (potential_neighbors.begin(), potential_neighbors.end(), f)
                         == potential_neighbors.end())
                        potential_neighbors.push_back(f);
                    else
                    {
                        neighbors.push_back(f);
                    }
                }
            }
        }
        faceNeighbors.push_back(neighbors);
    }
}



