/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CompSelection.h"
#include "MathTools.h"
#define distBinNb 10
#define angleBinNb 10

using namespace std;

/*
 * initialise the connected component selection
 * ie: compute the descriptor (approximate shape context) of each component
 * set all the distances and rank values to 0 (no reference component yet)
 */
void CompSelection::init(const Mesh &m){
    Trace(2, " CompSelection::init");
    Selection::init(m);
    QTime timer;
    if (SPY_TIME)
        timer.start();
    computeASCs();
    if (SPY_TIME)
        timing_precomputation.push_back(timer.elapsed());

    nb_of_similar_sel = 1;
}

 void CompSelection::saveTimings(const QString &filename){
     Trace(2, " CompSelection::saveTimings ");
     std::string file_n = qPrintable(filename);
     file_n.append("_comp");

     ofstream out (file_n.c_str ());
     out << "V: " << vertices.size() << " T: "<< triangles.size() << "\n";
     out << "\n";
     out << "------------------------------------- \n";
     out << "-------   precomputed --------------- \n";
     out << "------------------------------------- \n";

     for (unsigned int i = 0; i < timing_precomputation.size(); i++)
         out  << (float) timing_precomputation[i]<< "\n";

     out << "------------------------------------- \n";
     out << "-------     local    ---------------- \n";
     out << "------------------------------------- \n";
     int previous_sel = -1;
     for (unsigned int i = 0; i < timing_local.size(); i ++)
     {
         if ( (int) timing_local[i].first != previous_sel){
             out <<"\n" << timing_local[i].first<< " - ";
             previous_sel = timing_local[i].first;
         }
         out << (float) timing_local[i].second << " ";
     }

     out << "\n";
     out << "------------------------------------- \n";
     out << "-------    global    ---------------- \n";
     out << "------------------------------------- \n";

     for (unsigned int i = 0; i < timing_global.size(); i++)
         out <<"\n" << timing_global[i].first<< " - " << (float) timing_global[i].second << " ";
 }

QPixmap CompSelection::getUnderMouseHistogram(unsigned int v, float & dist){
    Trace(4, " CompSelection::getUnderMouseHistogram");
    unsigned int curSCind = vertexConnectedComp[v];
    dist = _shapeContexts[curSCind]._nDist;
    return _shapeContexts[curSCind]._shapeContext.drawHistogram();
}

void CompSelection::clear(){
    Trace(2, " CompSelection::clear");
    _shapeContexts.clear();
    timing_precomputation.clear();
    timing_global.clear();
    timing_local.clear();
}

void CompSelection::startSelection(UserStroke & stroke){
    Trace(2, "CompSelection::computeSelection ");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    std::vector<UserStroke::strokePoints> strokePoints = stroke.getPoints();
    assert (!strokePoints.empty());

    // use the last point of the stroke as a click
    // to find the selected component
    double d0, d1, d2;
    unsigned int triangle_under_point;
    barycentricCoordinates(strokePoints[strokePoints.size()-1].p,
                           triangle_under_point, d0, d1, d2);

    componentSelected = vertexConnectedComp[triangles[triangle_under_point].getVertex(0)];

    // set the selected faces
    unsigned int curInd;
    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        curInd = vertexConnectedComp[triangles[i].getVertex(0)];
        if (curInd == componentSelected) {
            faceInSelectionK[i] = 100.0f;
             Trace(4, " i " << i << " in selection ");
        }
        else
            faceInSelectionK[i] = -1.f;
    }
    // compute the distance between the reference ASC and others
    recomputeSimilarityDist();
}

void CompSelection::updateSelection(UserStroke & stroke){
    Trace(2, "CompSelection::refineSelection ");
    startSelection(stroke);
}

void CompSelection::endSelection(UserStroke & stroke){
    if(SPY_TIME)
        nb_of_local_sel++;
    Q_UNUSED(stroke);
}

void CompSelection::eraseSelection(UserStroke & stroke){
    Q_UNUSED(stroke);
}

void CompSelection::findSimilarSelections(UserStroke& stroke){
    Trace(2, "CompSelection::findSimilarSelections (Stroke)");
    QTime timer;
    if (SPY_TIME)
        timer.start();
    Q_UNUSED(stroke);
    simSelectionsNb = 0;

    findSimilarSelections();
    if (SPY_TIME){
        timing_global.push_back(make_pair( simSelectionsNb, timer.elapsed()));
    }
    simSelectionsNb++;
}

void CompSelection::findSimilarSelections(){
    Trace(2, " CompSelection::findSimilarSelections");
    unsigned int v;
    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        v = triangles[i].getVertex(0);
        if (faceInSelectionK[i] != 100.f )
       {
            if (_shapeContexts[vertexConnectedComp[v]]._rankToRef <= nb_of_similar_sel)
                faceInSelectionK[i] = 100.f + _shapeContexts[vertexConnectedComp[v]]._nDist+1.f;
            else
                faceInSelectionK[i] = -1.f;
        }
    }
}

void CompSelection::changeSimiliarityThreshold(int val){
    Trace(2, " CompSelection::changeSimilarityThreshold " << val);
    QTime timer;
    if (SPY_TIME)
        timer.start();
    nb_of_similar_sel =  val;
    findSimilarSelections();
    if (SPY_TIME){
        timing_global.push_back(make_pair( simSelectionsNb, timer.elapsed()));
    }
    simSelectionsNb++;
}

void CompSelection::clearSelections(){
    Trace(2, "CompSelection::clearSelections");

#pragma omp parallel for
    for (unsigned int i = 0; i<triangles.size(); i++)
        faceInSelectionK[i] = -1.f;

#pragma omp parallel for
    for (unsigned int i = 0; i < _shapeContexts.size(); i++)
    {
        _shapeContexts[i]._nDist = 0.f;
        _shapeContexts[i]._rankToRef = 0;
    }

}
void CompSelection::recomputeSimilarityDist(){
    Trace(2, " CompSelection::recomputeSimilarityDist");

    // compute the distances between the reference and other connected components
#pragma omp parallel for
    for (unsigned int i = 0; i < _shapeContexts.size(); i++)
    {
        switch (histoDist){
        case L2:
            _shapeContexts[i]._nDist =  _shapeContexts[componentSelected]._shapeContext.distanceL2(
                        _shapeContexts[i]._shapeContext);
            break;
        case L1:
            _shapeContexts[i]._nDist =  _shapeContexts[componentSelected]._shapeContext.distanceL1(
                        _shapeContexts[i]._shapeContext);
            break;
        case CHI2:
            _shapeContexts[i]._nDist =  _shapeContexts[componentSelected]._shapeContext.distanceChi2(
                        _shapeContexts[i]._shapeContext);
            break;
        case CHI2N:
            _shapeContexts[i]._nDist =  _shapeContexts[componentSelected]._shapeContext.distanceChi2Normalized(
                        _shapeContexts[i]._shapeContext);
            break;
        default:
            break;
        }
    }
    // sort the connected components by distance to the ref
    std::vector<std::pair<float, unsigned int> > sortToRef;
    for (unsigned int i = 0; i < _shapeContexts.size(); i++)
        sortToRef.push_back( make_pair(_shapeContexts[i]._nDist, i));
    sort(sortToRef.begin(), sortToRef.end());

#pragma omp parallel for
    for (unsigned i = 0 ; i < sortToRef.size(); i++){
        _shapeContexts[sortToRef[i].second]._rankToRef = i;
    }

    float max_min = sortToRef[sortToRef.size()-1].first - sortToRef[0].first; // max-min;
    if (max_min != 0){
        for (unsigned int i = 0; i < _shapeContexts.size(); i++)
            _shapeContexts[i]._nDist = (_shapeContexts[i]._nDist-sortToRef[0].first)/max_min;
    }
}

void CompSelection::computeASCs(){
    Trace(2,"CompSelection::computeASCs" );
    _shapeContexts.clear();
    componentShapeContext sc;
    for (unsigned int i = 0; i < connectedCompRepresentatives.size(); i++)
        _shapeContexts.push_back(sc);

#pragma omp parallel for
    for (unsigned int i = 0; i < connectedCompRepresentatives.size(); i++)
    {
        componentShapeContext cur_sc;
        cur_sc._indice = i;
        miniball(vertexConnectedComp, i, vertices, cur_sc._center, cur_sc._radiusMax, 0.02);
        computeASC(cur_sc);
        _shapeContexts[i] = cur_sc;
    }
}

void CompSelection::computeASC(componentShapeContext & curComp )
{
    Trace(4, "CompSelection::computeASC" );
    float dist;
    float angle;
    float area;
    Vec3Df r;
    curComp._shapeContext = Hist2Df(distBinNb, angleBinNb, 0, 1, 0, M_PI);
    for (unsigned int i = 0; i < vertices.size(); i++)
    {
        if (vertexConnectedComp[i] == curComp._indice)
        {
            r = curComp._center-vertices[i].getPos();
            dist = r.getLength();
            r.normalize();
            Vec3Df n_i = vertices[i].getSmoothNormal();
            angle = acos(Vec3Df::dotProduct(r,n_i));
            area  = 0.f;
            const std::vector<unsigned int> & faceV = vertices[i].getFaces();
            for (unsigned int j = 0; j < faceV.size(); j++)
                area += triangleArea(faceV[j])/3.f ;
            curComp._shapeContext.addValue(dist/curComp._radiusMax, angle, area );
        }
    }
//    curComp._shapeContext.normalize();
}

void CompSelection::drawShapeContext(unsigned int i){
    Vec3Df center = _shapeContexts[i]._center;
//    float diskRadius = _shapeContexts[i]._radiusMax;

    glColor4f(0.9f, 0.2f, 0.2f, 1.f);
    BasicGL::drawSphere(center[0], center[1], center[2], 0.005, 20, 20);

//    glColor4f(0.9f, 0.2f, 0.2f, 0.5f);
//    BasicGL::drawSphere(center[0], center[1], center[2], diskRadius, 20, 20);
//    glColor4f(0.0f, 1.f, 0.2f, 0.5f);
//    BasicGL::drawSphere(refCenter[0], refCenter[1], refCenter[2], 0.05, 15, 15);
}

