/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "PartSelection.h"

#define CONSTRAINT_WEIGHT 12000.f
#define ISOLINES_NB 15
#define HIST_SIZE 20
#include <queue>

using namespace std;

/**
 * @brief The isoVal struct
 */
struct isoVal{
    unsigned int f;
    unsigned int v_i;
    unsigned int v_j;
    float t;
};

void PartSelection::init(const Mesh &m)
{
    Trace(2, "PartSelection::init");

    Selection::init(m);
    histoDist = CHI2;

    QTime timer;
    if (SPY_TIME)
        timer.start();

    _cotanLaplacian = new CotanLaplacian;
    _cotanLaplacian->init(m);
    _cotanLaplacian->computeLaplacian();

    // here we tried at some point to use other types of laplacians
    _laplacian = _cotanLaplacian;

    userDefinedHarmonicField.clear();
    for (unsigned int i = 0; i < vertices.size(); i++)
        userDefinedHarmonicField.push_back(0.f);

    computeConformalFactor();
    cholmodInitialized = false;
    initCholmod(false, constraints0, constraints1);

    simSelectionsNb = 0;
    nb_of_similar_sel = 1;

    if (SPY_TIME)
        timing_precomputation.push_back(timer.elapsed());

    numberOfStrokePnts = 0;
    numberOfRefineStrokePnts = 0;
    numberOfBackGroundPnts = 0;
}

void PartSelection::clear(){
    Trace(4, "PartSelection::clear");
    // user stroke
    numberOfStrokePnts = 0;
    numberOfRefineStrokePnts = 0;
    numberOfBackGroundPnts = 0;

    // selection
    constraints0.clear();
    constraints1.clear();

    userDefinedHarmonicField.clear();
    normalizedHarmonicField.clear();
    if (cholmodInitialized)
        cholmod_struct.free();

    // expansion
    conformalFactor.clear();
    for (unsigned int i = 0; i < potentialStrokes.size(); i++)
        potentialStrokes[i].clear();
    potentialStrokes.clear();

    // timing
    timing_precomputation.clear();
    timing_global.clear();
    timing_local.clear();
}

void PartSelection::saveTimings(const QString &filename){
    std::string file_n = qPrintable(filename);
    file_n.append("_part");

    ofstream out (file_n.c_str ());
    out << "V: " << vertices.size() << " T: "<< triangles.size() << "\n";
    out << "\n";
    out << "------------------------------------- \n";
    out << "-------   precomputed --------------- \n";
    out << "------------------------------------- \n";

    for (unsigned int i = 0; i < timing_precomputation.size(); i++)
    {
            out  << (float) timing_precomputation[i]<< "\n";
    }

    out << "------------------------------------- \n";
    out << "-------     local    ---------------- \n";
    out << "------------------------------------- \n";
    int previous_sel = -1;
    for (unsigned int i = 0; i < timing_local.size(); i ++)
    {
        if ( (int) timing_local[i].first != previous_sel){
            out <<"\n" << timing_local[i].first<< " - ";
            previous_sel = timing_local[i].first;
        }
        out << (float) timing_local[i].second << " ";
    }

    out << "\n";
    out << "------------------------------------- \n";
    out << "-------    global    ---------------- \n";
    out << "------------------------------------- \n";

    for (unsigned int i = 0; i < timing_global.size(); i++)
    {
            out <<"\n" << timing_global[i].first<< " - " << (float) timing_global[i].second << " \n ";
            for (unsigned int j = 0; j < timing_by_part[i].size(); j++)
            {
                out << "     " << timing_by_part[i][j].first << " - " << timing_by_part[i][j].second << "\n";
            }
    }
}
//***************************************************
//                  SELECTION
//***************************************************

void PartSelection::startSelection(UserStroke & stroke){
    QTime timer;
    if (SPY_TIME)
        timer.start();

    numberOfStrokePnts = 0;
    Trace(2, "PartSelection::computeSelection ");

    // we clean the linear systema from previous user constraints --> cholmod downdate
    setCholmodUpdate(false, constraints0, constraints1);
    constraints0.clear();
    constraints1.clear();

    // we find the new constraints from the user stroke
    vector<UserStroke::strokePoints> strokePoints = stroke.getPoints();
    setHarmonicConstraints(strokePoints.size(), strokePoints, constraints0, constraints1);
    numberOfStrokePnts = strokePoints.size();

    // and we update the linear systema with this new constraints
    if (! constraints0.empty())
    {
        setCholmodUpdate(true, constraints0, constraints1);
        findIsolines(stroke);
        findBestIsoline();
        findSelectedFaces();
    }

    if (SPY_TIME)
        timing_local.push_back(make_pair( nb_of_local_sel, timer.elapsed()));
}


void PartSelection::updateSelection(UserStroke & stroke){
    Trace(4, "PartSelection::updateSelection ");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    // we find the new constraints from the user stroke and add them to the previous ones
    vector<UserStroke::strokePoints> strokePoints = stroke.getPoints(); // stroke.getLastStroke();
    if (! strokePoints.empty())
    {
        // caution! it doesn't work if i dont downdate and then update with everything...
        // first we clean the linear systema from previous user constraints
        // --> cholmod downdate
        setCholmodUpdate(false, constraints0, constraints1);

        vector<unsigned int> newConstraints0;
        vector<unsigned int> newConstraints1;
        setHarmonicConstraints(strokePoints.size(), strokePoints, newConstraints0, newConstraints1);

        constraints0 = newConstraints0;
        constraints1 = newConstraints1;

        // and we update the linear systema with all this constraints
        setCholmodUpdate(true, constraints0, constraints1);

        findIsolines(stroke);
        findBestIsoline();
        findSelectedFaces();
        numberOfStrokePnts = strokePoints.size();
    }

    if (SPY_TIME)
        timing_local.push_back(make_pair( nb_of_local_sel, timer.elapsed()));
}


void PartSelection::endSelection(UserStroke & stroke){
    Q_UNUSED(stroke);
    nb_of_local_sel++;
}

void PartSelection::eraseSelection(UserStroke & stroke){
    vector<UserStroke::strokePoints> bgPoints = stroke.getBGPoints();

    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    for (unsigned int i = numberOfBackGroundPnts ; i < bgPoints.size(); i++)
    {
        findTrianglesInsideRadius(bgPoints[i].p, bgPoints[i].r, viewedFaces, 2, 6);
    }
    for (unsigned int i = 0; i < triangles.size(); i++)
        if (viewedFaces[i] == 6 )
            faceInSelectionK[i] = -1.f;

    numberOfBackGroundPnts = bgPoints.size();
}


void PartSelection::refineSelection(UserStroke& _stroke){
    vector< UserStroke::strokePoints > points = _stroke.getPoints();

    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    for (unsigned int i = numberOfRefineStrokePnts ; i < points.size(); i++)
    {
        findTrianglesInsideRadius(points[i].p, points[i].r, viewedFaces, 2, 6);
    }
    float colorValue = -1;
    for (unsigned int i = 0; i < triangles.size(); i++){
        if (viewedFaces[i] == 6 && faceInSelectionK[i] != -1)
            colorValue = faceInSelectionK[i];
    }
    if (colorValue != -1){
    for (unsigned int i = 0; i < triangles.size(); i++)
        if (viewedFaces[i] == 6 )
            faceInSelectionK[i] = colorValue;
    }
    numberOfRefineStrokePnts = points.size();
}

void PartSelection::clearSelections(){
    for (unsigned int i = 0; i<triangles.size(); i++)
        faceInSelectionK[i] = -1.f;
    simSelectionsNb = 0;
    refHist.clearValues();
    potentialSelections.clear();
    potentialStrokes.clear();
    for (unsigned int i = 0; i  < isolines.size(); i++)
        isolines[i].clear();
    isolines.clear();
}



bool PartSelection::findTrianglesInsideRadius( Vec3Df point, float radius,
                                                std::vector<int> &  inside,
                                                int inside_value,
                                                int new_inside_value)
{
    bool found = false;
    unsigned int i, nei_f, cur_f;
    Vec3Df c;
    queue< unsigned int > toClassify;
    float dist ;

    i = 0;
    while ( !(found || i >=triangles.size() ))
    {
        c = getBar(i);
        dist = (c-point).getLength();
        if ( (dist<=radius) && inside[i] != inside_value)
            found = true;
        else
            i++;
    }
    if (found)
    {
        inside[i] = new_inside_value;
        for (unsigned int j = 0; j < faceNeighbors[i].size(); j++)
        {
            nei_f = faceNeighbors[i][j];
            if (inside[nei_f] != new_inside_value && inside[nei_f] != inside_value && inside[nei_f] != 4)
            {
                toClassify.push(nei_f);
                inside[nei_f] = 4;
            }
        }
        while (!toClassify.empty())
        {
            cur_f = toClassify.front();
            toClassify.pop();
            c = getBar(cur_f);
            dist = (c-point).getLength();
            if (dist < radius)
            {
                inside[cur_f] = new_inside_value;
                for (unsigned int f = 0; f < faceNeighbors[cur_f].size(); f++)
                {
                    nei_f = faceNeighbors[cur_f][f];
                    if (inside[nei_f] != new_inside_value && inside[nei_f] != 4)
                    {
                        toClassify.push(nei_f);
                        inside[nei_f] = 4;
                    }
                }
            } else
                 inside[cur_f] = 3;
        }
    }
    return found;
}

//***************************************************
//                  EXPANSION
//***************************************************


void PartSelection::findSimilarSelections(UserStroke& stroke){
    Trace(2, "PartSelection::findSimilarSelections ");

    QTime timer;
    if (SPY_TIME)
        timer.start();
    simSelectionsNb = 0;

    vector<UserStroke::strokePoints> strokeInit = stroke.getPoints();

    // find the isolines with the same average conformal factor of the stroke.
    findConformalIsolines(strokeInit, potentialStrokes);

    // save the first set of selected faces, and compute its conformal factor average
    vector<bool> selectedVertices;
    for (unsigned int i = 0; i < vertices.size(); i++)
        selectedVertices.push_back(false);

    refHist.setBounds(1.f, 0.f);
    refHist.setBinsNumber(HIST_SIZE);
    unsigned int v;
    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        if (faceInSelectionK[i] == 1000.f)
        {
            for (unsigned int j = 0; j < 3; j++)
            {
                v = triangles[i].getVertex(j);
                if ( !selectedVertices[v])
                {
                    selectedVertices[v] = true;
                    float verticeArea = 0.f;
                    for (unsigned int k = 0; k < oneRingTr[v].size(); k++ )
                        verticeArea += triangleArea(oneRingTr[v][k])/3.f;
                    refHist.addValue(conformalFactor[v], verticeArea);
                }
            }
        }
    }

    // Cholmod downdate
    simSelectionsNb = 0;
    setCholmodUpdate(false, constraints0, constraints1);
    std::vector<unsigned int> _constraints0;
    std::vector<unsigned int> _constraints1;

    potentialSelections.clear();
    std::vector< std::pair<unsigned int, float> > loc_timing_by_part;
    Trace(1, " Number of potential Selections " << potentialStrokes.size());

    for (unsigned int i = 0; i < potentialStrokes.size(); i++)
    {
        QTime timer2;
        if (SPY_TIME)
            timer2.start();

        if (  potentialStrokes[i].size() > 2 )
        {
            setHarmonicConstraints( strokeInit.size() , potentialStrokes[i] ,
                                    _constraints0, _constraints1);

            // compute the field for the current potential stroke
            setCholmodUpdate(true, _constraints0, _constraints1);

            // compute the potential selection (fill the curSelection struct..)
            potentialSelection curSelection;
            setPotentialSelection(curSelection);
            curSelection.c0 = _constraints0;
            curSelection.c1 = _constraints1;
            potentialSelections.push_back(curSelection);
            // cholmod downdate
            setCholmodUpdate(false, _constraints0, _constraints1);
            _constraints0.clear();
            _constraints1.clear();
        }
        if (SPY_TIME){
            loc_timing_by_part.push_back( make_pair( i, timer2.elapsed()));
        }
    }
    timing_by_part.push_back(loc_timing_by_part);

    // classify all the potential selection ad set their rank
    vector<pair<float, unsigned int> > rangedSelections;
    for (unsigned int i = 0; i < potentialSelections.size(); i++)
        rangedSelections.push_back(make_pair(potentialSelections[i].dist_to_ref, i));
    sort(rangedSelections.begin(), rangedSelections.end());
    for (unsigned int i = 0; i< rangedSelections.size(); i++)
    {
        potentialSelections[rangedSelections[i].second].range = i;
        findMultipleSelectedFaces(potentialSelections[rangedSelections[i].second]);
    }

    constraints0 = _constraints0;
    constraints1 = _constraints1;

    if (SPY_TIME){
        timing_global.push_back(make_pair( simSelectionsNb, timer.elapsed()));
    }
    simSelectionsNb++;
}

void PartSelection::changeSimiliarityThreshold(int val){
    Trace(4, "PartSelection::changeSimilarityThreshold");

    nb_of_similar_sel = val;
    for (unsigned int i = 0; i < triangles.size(); i++)
            if ( faceInSelectionK[i] > 1000.f)
                faceInSelectionK[i] = -1.f;

    for (unsigned int i = 0; i< potentialSelections.size(); i++)
    {
        findMultipleSelectedFaces(potentialSelections[i]);
    }
}

const std::vector<float> & PartSelection::getUserField(){
    float min = FLT_MAX;
    float max = FLT_MIN;
    for (unsigned int i = 0; i < userDefinedHarmonicField.size(); i ++)
    {
        if (userDefinedHarmonicField[i] > max)
            max = userDefinedHarmonicField[i];
        if (userDefinedHarmonicField[i] < min)
            min = userDefinedHarmonicField[i];
    }
    float max_min = max-min;
    normalizedHarmonicField.clear();
    if (max_min != 0)
    {
        for (unsigned int i = 0; i < userDefinedHarmonicField.size(); i++)
            normalizedHarmonicField.push_back( (userDefinedHarmonicField[i]-min)/max_min);
        return normalizedHarmonicField;
    }
    else {
        return userDefinedHarmonicField;
    }
}


//***************************************************
//                  ISOLINE STUFFS
//***************************************************

void PartSelection::findIsolines(UserStroke & curStroke){

    Trace(2, "PartSelection::findIsolines");
    for (unsigned int i = 0; i < isolines.size(); i++)
        isolines[i].clear();
    isolines.clear();
    for (unsigned int j = 0; j < ISOLINES_NB; j ++)
    {
        vector< vector< Vec3Df > > isoLines_j;
        float iso = (float) j/(float)(ISOLINES_NB+1);
        findIsoline(iso, userDefinedHarmonicField, isoLines_j);

        if (!isoLines_j.empty())
        {
            unsigned int isoLineToAdd;
            if (isoLines_j.size() != 1)
                isoLineToAdd = findIsolineToKeep(curStroke, isoLines_j);
            else
                isoLineToAdd = 0;
            isolines.push_back(isoLines_j[isoLineToAdd]);
        } else {
            Trace(4, " error isoline "<< j << " not found !! \n");
            vector<Vec3Df> empty_line;
            isolines.push_back(empty_line);
        }
    }
    Trace(4, " i0solines[0] size " << isolines[0].size());
}

void PartSelection::findNextIsolineElement( std::vector<Vec3Df> & isoline,
                                            std::vector<bool> & triangleViewed,
                                            const std::vector<unsigned int> & triangleInd,
                                            const std::vector<float> & isoValues,
                                            const float refIsoVal,
                                            unsigned int v0,
                                            unsigned int v1,
                                            unsigned int t)
{
    std::vector<unsigned int> neighbors = faceNeighbors[t];

    unsigned int tr;
    for (unsigned int i = 0; i < neighbors.size(); i++)
    {
        tr = neighbors[i];
        if (triangles[tr].contains(v0) && triangles[tr].contains(v1) && !triangleViewed[tr])
        {
            unsigned int v_not_in_t = 4;
            for (unsigned int j = 0; j < 3; j++){
                if ( ! triangles[t].contains(triangles[tr].getVertex(j)))
                {
                    v_not_in_t = j;
                    break;
                }
            }
            assert(v_not_in_t != 4);

            unsigned int val = triangleInd[tr];
            unsigned int v2, v3;
            v2 = vertices.size();
            v3 = vertices.size();
            switch (val){
            case 0 : case 7:
                break;
            case 1: case 6:
                if (v_not_in_t == 1)
                {
                    v2 = triangles[tr].getVertex(0);
                    v3 = triangles[tr].getVertex(1);
                } else {
                    v2 = triangles[tr].getVertex(0);
                    v3 = triangles[tr].getVertex(2);
                }
                break;
            case 2 : case 5:
                if (v_not_in_t == 0)
                {
                    v2 = triangles[tr].getVertex(0);
                    v3 = triangles[tr].getVertex(1);
                } else {
                    v2 = triangles[tr].getVertex(2);
                    v3 = triangles[tr].getVertex(1);
                }
                break;
            case 3 : case 4:
                if (v_not_in_t == 1)
                {
                    v2 = triangles[tr].getVertex(1);
                    v3 = triangles[tr].getVertex(2);
                } else {
                    v2 = triangles[tr].getVertex(0);
                    v3 = triangles[tr].getVertex(2);
                }
                break;
            }
            assert (v2 != vertices.size() && v3 != vertices.size() );

            float isoT = (refIsoVal-isoValues[v2])/(isoValues[v3]-isoValues[v2]);
            triangleViewed[tr] = true;
            isoline.push_back( vertices[v2].getPos()+isoT*(vertices[v3].getPos()-vertices[v2].getPos()));
            findNextIsolineElement(isoline, triangleViewed, triangleInd, isoValues, refIsoVal, v2, v3, tr);
            break;
        }
    }
}

void PartSelection::findIsoline(float refIsoVal, const std::vector<float> & isoValues,
                                std::vector< std::vector<Vec3Df> > & isoLines_i )
{
    Trace(3, "PartSelection::findIsoline refIsoVal");
    std::vector<unsigned int> triangleInd;
    std::vector<bool> triangleViewed;
    std::vector<unsigned int> onIsoline;

    unsigned int val;

    for (unsigned int i =0; i < triangles.size(); i++)
    {
        val = 0;
        /* assigned a value for each triangle depending on the isovalues of it vertices
        // 0 if a vertex isovalue is below the isoline, 1 if it is above
        //   1:         2:         3:            |       0:
        //                                       |
        //   1          0          1             |       0
        //  / \        / \        / \            |      / \
        // 0 - 0      0 - 1      0 - 1           |     0 - 0
        //
        //   6:         5:         4:            |       6:
        //                                       |
        //   0          1          0             |       1
        //  / \        / \        / \            |      / \
        // 1 - 1      1 - 0      1 - 0           |     1 - 1 */

        for (int j = 2; j >= 0; j--)
        {
            if ( isoValues[triangles[i].getVertex(j)] > refIsoVal)
                val |= (1<<j);

        }
        triangleInd.push_back(val);
        triangleViewed.push_back(false);
        if (val < 7 && val > 0)
            onIsoline.push_back(i);
    }

    std::vector<Vec3Df> isoCurL;
    std::vector<Vec3Df> isoCurR;
    // while all the triangles on the isoline are not treated
    // we take on triangle on the isoline and retrieve the isoline that goes through it
    while (! onIsoline.empty())
    {
        unsigned int tr = onIsoline.back();
        onIsoline.pop_back();
        triangleViewed[tr] = true;
        unsigned int val = triangleInd[tr];
        unsigned int v0= vertices.size(), v1= vertices.size(), v2= vertices.size(), v3= vertices.size();
        isoCurL.clear();
        isoCurR.clear();
        switch (val){
        case 0 : case 7:
            v0 = vertices.size();
            v1 = vertices.size();
            v2 = vertices.size();
            v3 = vertices.size();
            break;
        case 1 : case 6:
            v0 = triangles[tr].getVertex(0);
            v1 = triangles[tr].getVertex(1);
            v2 = triangles[tr].getVertex(0);
            v3 = triangles[tr].getVertex(2);
            break;
        case 2 : case 5:
            v0 = triangles[tr].getVertex(0);
            v1 = triangles[tr].getVertex(1);
            v2 = triangles[tr].getVertex(2);
            v3 = triangles[tr].getVertex(1);
            break;
        case 3: case 4:
            v0 = triangles[tr].getVertex(0);
            v1 = triangles[tr].getVertex(2);
            v2 = triangles[tr].getVertex(1);
            v3 = triangles[tr].getVertex(2);
            break;
        }
        // we actually have to take care about the two direction of the isoline
        // (going out from triangle t)
        // to handle correctly models with boundaries
        assert (v0 != vertices.size() && v1 != vertices.size() );
        float t = (refIsoVal-isoValues[v0])/(isoValues[v1]-isoValues[v0]);
        isoCurL.push_back(vertices[v0].getPos()+t*(vertices[v1].getPos()-vertices[v0].getPos()));
        findNextIsolineElement(isoCurL, triangleViewed, triangleInd, isoValues, refIsoVal, v0, v1, tr);

        assert (v2 != vertices.size() && v3 != vertices.size() );
        t = (refIsoVal-isoValues[v2])/(isoValues[v3]-isoValues[v2]);
        isoCurR.push_back(vertices[v2].getPos()+t*(vertices[v3].getPos()-vertices[v2].getPos()));
        findNextIsolineElement(isoCurR, triangleViewed, triangleInd, isoValues, refIsoVal, v2, v3, tr);
        if (! isoCurR.empty() && isoCurR.size() < isoCurL.size())
        {
            std::reverse(isoCurR.begin(), isoCurR.end());
            isoCurR.insert(isoCurR.end(), isoCurL.begin(), isoCurL.end());
            isoLines_i.push_back(isoCurR);
        }
        else if (! isoCurL.empty() && isoCurL.size() < isoCurR.size())
        {
            std::reverse(isoCurL.begin(), isoCurL.end());
            isoCurL.insert( isoCurL.end(), isoCurR.begin(), isoCurR.end());
            isoLines_i.push_back(isoCurL);
        } else if (isoCurL.empty() && !isoCurR.empty())
            isoLines_i.push_back(isoCurR);
        else if (isoCurR.empty() && !isoCurL.empty())
                   isoLines_i.push_back(isoCurL);

        for (int i  = onIsoline.size()-1; i >= 0 ; i--)
            if (triangleViewed[onIsoline[i]])
                onIsoline.erase(onIsoline.begin()+i);
    }
}

unsigned int PartSelection::findIsolineToKeep(UserStroke & curStroke,
                                              std::vector< std::vector<Vec3Df> > & isoLines_i)
{

    Trace(3, "PartSelection::findIsolineToKeep");
    std::vector<UserStroke::strokePoints> stroke = curStroke.getPoints();
    // we need to keep the isoline which go through the user stroke. ( ie : the closest one )
    // we only check distance with the median point of the stroke.
    // as the other potential isolines with the same isovalue should be far away
    Vec3Df middlePoint = stroke[stroke.size()/2].p;
    float min = FLT_MAX;
    float min_by_isoline;
    unsigned int isolineSelected = 0;
    for (unsigned int i = 0; i <isoLines_i.size(); i++)
    {
        min_by_isoline = FLT_MAX;
        for (unsigned int j = 0; j < isoLines_i[i].size(); j++)
        {
            float d = (isoLines_i[i][j]-middlePoint).getLength();
            if ( d < min_by_isoline)
                min_by_isoline = d;
        }
        if (min_by_isoline < min)
        {
            min = min_by_isoline;
            isolineSelected = i;
        }
    }
    return isolineSelected;
}


void PartSelection::findBestIsoline()
{
    Trace(2, "PartSelection::findBestIsoline");
    // compute the radius of each isoline.
    vector<float> isolines_radius;
    for (unsigned int i = 0; i < isolines.size(); i++)
        isolines_radius.push_back(0.f);
    Vec3Df d;
    for (unsigned int i = 0; i < isolines.size(); i++)
    {
        for (unsigned int j = 1; j < isolines[i].size(); j++)
        {
            d = isolines[i][j]-isolines[i][j-1];
            isolines_radius[i]+= d.getLength();
        }
        if (isolines[i].size()>1)
            isolines_radius[i]+= (isolines[i][0]-isolines[i][isolines[i].size()]).getLength();
        else
            isolines_radius[i] = FLT_MAX;
        isolines_radius[i]/= (2*M_PI);
        Trace(4, " isoline " << i << " radius " << isolines_radius[i]);
    }

    // compute delta_i for each isoline. ref (Mesh decomposition with cross boundary brushes).
    unsigned int k_max;
    float f_k_sum, f_k_delta_k, f_k;
    vector<float> delta_i;
    float t = isolines.size()/2.f;
    unsigned int nb_isolines = isolines.size();
    for (unsigned int i = 0; i < isolines.size(); i++)
    {
        k_max = min(i, nb_isolines-i);
        f_k_sum = 0.f;
        f_k_delta_k = 0.f;
        for (unsigned int k = 1; k < k_max; k++)
        {
            f_k = exp(- ((k-1.f)*(k-1.f))/8.f);
            f_k_sum += f_k;
            f_k_delta_k += f_k*(2*isolines_radius[i]-isolines_radius[i-k]-isolines_radius[i+k]);
        }
        delta_i.push_back(f_k_delta_k/f_k_sum);
        delta_i[i]*= exp(-((i-t)*(i-t)/(2*t*t)));
    }
    float m_min = FLT_MAX;
    for (unsigned int i = 0; i < delta_i.size(); i++)
    {
        if (delta_i[i] < m_min)
        {
             m_min = delta_i[i];
             refIsoline = i;
        }
    }
    Trace(3, "refIsoline " << refIsoline);
}


//***************************************************
//                  EXPANSION
//***************************************************


void PartSelection::computeConformalFactor(){
    Trace(2, "PartSelection::computeConformalFactor");
    vector<float> gaussianCurvature;
    computeGaussianCurvature(gaussianCurvature);

    float sum = 0.f;
    float verticeArea;
    for (unsigned int i = 0; i < vertices.size(); i++ )
    {
        verticeArea = 0.f;
        for (unsigned int j = 0; j < oneRingTr[i].size(); j++ )
            verticeArea += triangleArea(oneRingTr[i][j])/3.f;
        sum += gaussianCurvature[i]*verticeArea;
    }

    float sumArea = 0.f;
    for (unsigned int i = 0; i < triangles.size(); i++)
        sumArea += triangleArea(i);

    sum /= sumArea;

    vector<float> Korigin;
    for (unsigned int i = 0 ; i < vertices.size(); i ++)
        Korigin.push_back(sum);

    cholmod_struct.start();

    // Matrice A = [ L WP]T, #lines = n + 2c, #columns = n
    cholmod_struct.setDimensions(vertices.size()+connectedCompRepresentatives.size(), vertices.size(), 1);

    //Number of nnz in L
    int _nb_non_zeros_in_A = vertices.size();
    for( unsigned int i = 0 ; i < oneRing.size(); i++ ){
        _nb_non_zeros_in_A += oneRing[i].size();
    }
    // the last line is full of one ( to compute the mean of the conformal factor)
    _nb_non_zeros_in_A += connectedCompRepresentatives.size(); //nbOfConstraints;

    cholmod_struct.specifyNonZeroInA(_nb_non_zeros_in_A);
    cholmod_struct.allocate();

    // SET DE LA MATRICE A DANS CHOLMOD
    // first, we enter all the values of the laplacian
    const vector< vector<pair <unsigned int, float> > > laplacianValues = _cotanLaplacian->getLaplacian();

    for( unsigned int i = 0; i < laplacianValues.size() ; i++ )
        for( unsigned int j = 0; j < laplacianValues[i].size(); j++ )
            cholmod_struct.addValueInA(i, laplacianValues[i][j].first, laplacianValues[i][j].second );

    // we add a constraint in A for each connected component representative
    for (unsigned int i = 0 ; i < connectedCompRepresentatives.size(); i ++)
        cholmod_struct.addValueInA(vertices.size()+i, connectedCompRepresentatives[i], CONSTRAINT_WEIGHT);

    Trace(1, "--------before factorizing A in conformal factor");
    cholmod_struct.factorize();

    Trace(1, "--------after factorizing A in conformal factor");

    for (unsigned int i = 0; i < vertices.size(); i ++)
        cholmod_struct.setValueInB(i, 0, gaussianCurvature[i]-Korigin[i]);

     for (unsigned int i = 0 ; i < connectedCompRepresentatives.size(); i ++)
         cholmod_struct.setValueInB(vertices.size()+i, 0, 0.f);

    cholmod_struct.solve(false);

    conformalFactor.clear();
    float min = FLT_MAX;
    float max = FLT_MIN;
    float res;
    for (unsigned int i = 0; i < vertices.size(); i++)
    {
        res = cholmod_struct.getSolutionValue(i,0);
        conformalFactor.push_back(res );
        if (res > max)
            max = res;
        if (res < min)
            min = res;
    }

    // normalization
    float max_min = max-min;
    for (unsigned int i = 0; i < conformalFactor.size(); i++)
        conformalFactor[i] = (conformalFactor[i]-min)/max_min;

    cholmod_struct.freeSolution();
    cholmod_struct.free();
}

void PartSelection::findConformalIsolines( const std::vector<UserStroke::strokePoints> & strokeInit,
                                           std::vector< std::vector<Vec3Df> >  & conf_iso_lines )
{
    Trace(2, "PartSelection::findConformalIsolines");
    float refValue = 0.f;
    unsigned int t;
    double d0, d1, d2;
    // compute the reference value of the user stroke
    for (unsigned int i = 0; i < strokeInit.size(); i++)
    {
        barycentricCoordinates(strokeInit[i].p, t, d0, d1, d2);
        refValue += d0*conformalFactor[triangles[t].getVertex(0)]+
                d1*conformalFactor[triangles[t].getVertex(1)]+
                d2*conformalFactor[triangles[t].getVertex(2)];

    }
    refValue/=(float) strokeInit.size();
    findIsoline(refValue, conformalFactor, conf_iso_lines);
}

bool PartSelection::findMultipleSelectedFaces(const potentialSelection & curSelection)
{
    Trace(2, "  PartSelection::findMultipleSelectedFaces ");
    if (curSelection.range <= nb_of_similar_sel)
    {
        for (unsigned int i = 0; i < curSelection.faces.size(); i++)
            if (faceInSelectionK[curSelection.faces[i]] == 1000.f)
            {
                Trace(1, " return false in find multiple selected faces");
                return false;
            }
        for (unsigned int i = 0; i < curSelection.faces.size(); i++)
            faceInSelectionK[curSelection.faces[i]] = 1000.f+ curSelection.dist_to_ref;
        simSelectionsNb++;
        return true;
    }
    return false;
}


void PartSelection::setPotentialSelection( potentialSelection & curSel )
{
    Trace(2, "PartSelection::setPotentialSelection");

    // histogram initialization
    Hist1Df cur_hist;
    cur_hist.setBounds(1.f, 0.f);
    cur_hist.setBinsNumber(HIST_SIZE);
    cur_hist.clearValues();

    double r;
    vector<unsigned int> selected_f;
    vector<bool> selected_v;

    for (unsigned int i = 0; i < vertices.size(); i++)
        selected_v.push_back(false);

    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        const Triangle &t = triangles[i];
        unsigned int j = 0;
        bool triangleSelected = false;
        while ( !triangleSelected && j < 3)
        {
            r = cholmod_struct.getSolutionValue(t.getVertex(j), 0);
            if( r >  (float) refIsoline/(float)(ISOLINES_NB+1) )
            {
                selected_f.push_back(i);
                triangleSelected = true;
                for (unsigned int v_ind = 0; v_ind < 3; v_ind++)
                {
                    unsigned int v = triangles[i].getVertex(v_ind);
                    if ( !selected_v[v] )
                    {
                        selected_v[v] = true;
                        float verticeArea = 0.f;
                        for (unsigned int k = 0; k < oneRingTr[v].size(); k++ )
                            verticeArea += triangleArea(oneRingTr[v][k])/3.f;
                        cur_hist.addValue(conformalFactor[v], verticeArea);
                    }
                }
            }
            j++;
        }
    }
    float dist;
    switch (histoDist){
    case L2:
        dist = refHist.distanceL2(cur_hist);
        break;
    case L1:
         dist = refHist.distanceL1(cur_hist);
        break;
    case CHI2:
        dist = refHist.distanceChi2(cur_hist);
        break;
    case CHI2N:
        dist = refHist.distanceChi2Normalized(cur_hist);
        break;
    default:
        dist = 0.f;
        break;
    }
    curSel.dist_to_ref = dist;
    curSel.faces = selected_f;
    curSel.hist = cur_hist;
}


void PartSelection::recomputeSimilarityDist(){

#pragma omp parallel for
    for (unsigned int i = 0; i < potentialSelections.size(); i++)
    {
        switch (histoDist){
        case L2:
            potentialSelections[i].dist_to_ref = refHist.distanceL2(potentialSelections[i].hist);
            break;
        case L1:
            potentialSelections[i].dist_to_ref = refHist.distanceL1(potentialSelections[i].hist);
            break;
        case CHI2:
            potentialSelections[i].dist_to_ref = refHist.distanceChi2(potentialSelections[i].hist);
            break;
        case CHI2N:
            potentialSelections[i].dist_to_ref = refHist.distanceChi2Normalized(potentialSelections[i].hist);
            break;
        default:
            potentialSelections[i].dist_to_ref = 0.f;
            break;
        }

    }
    if (! potentialSelections.empty())
    {
        for (unsigned int i = 0; i < triangles.size(); i++)
            if ( faceInSelectionK[i] > 1000.f)
                faceInSelectionK[i] = -1.f;
    }

    vector<pair<float, unsigned int> > rangedSelections;
    for (unsigned int i = 0; i < potentialSelections.size(); i++)
        rangedSelections.push_back(make_pair(potentialSelections[i].dist_to_ref, i));
    sort(rangedSelections.begin(), rangedSelections.end());
    for (unsigned int i = 0; i< rangedSelections.size(); i++)
    {
        potentialSelections[rangedSelections[i].second].range = i;
        findMultipleSelectedFaces(potentialSelections[rangedSelections[i].second]);
    }
}

//***************************************************
//      FNS FOR THE COMPUTATION OF HARMONIC FIELDS
//***************************************************

void PartSelection::setHarmonicConstraints( unsigned int nb_of_ref_points,
                                            const std::vector<Vec3Df> & strokePoints,
                                             std::vector<unsigned int> & constraints0,
                                             std::vector<unsigned int> & constraints1 )
{
    Trace(2, "PartSelection::setHarmonicConstraints");
    Vec3Df lastPointIn(FLT_MIN, FLT_MIN, FLT_MIN);

    vector <unsigned int> v0vector;
    vector <unsigned int> v1vector;
    int v0vsv1 = 0;

    unsigned int space_bw_pnts = strokePoints.size()/nb_of_ref_points;

    for (unsigned int i = 0; i < strokePoints.size(); i+=space_bw_pnts)
    {
        if (lastPointIn == Vec3Df(FLT_MIN, FLT_MIN, FLT_MIN))
            lastPointIn = strokePoints[i];
        else
        {
            Vec3Df p_i = (lastPointIn+strokePoints[i])/2.f;
            unsigned int t_i = triangleCloseToPoint(p_i, 0.3);
            Vec3Df n_t_i = triangles[t_i].getNorm();
            Vec3Df n_i = strokePoints[i]-lastPointIn;
            n_i.normalize();
            Vec3Df v_dir = Vec3Df::crossProduct(n_i, n_t_i);
            v_dir.normalize();
            Vec3Df p1, p2;
            p1 =  p_i+0.01*v_dir; // todo try to change that by the radius when we have the user stroke ?
            p2 =  p_i-0.01*v_dir;
            int v1 = vertexCloseToPoint(p1, 0.1);
            int v2 = vertexCloseToPoint(p2, 0.1);
            // todo : something smarter than conformal factor to find R/L
            if (conformalFactor[v1] < conformalFactor[v2])
                v0vsv1--;
            else
                v0vsv1++;

            if (v1 != -1 && v2 != -1)
            {
                if ( find(v0vector.begin(), v0vector.end(), v1) == v0vector.end())
                    v0vector.push_back(v1);
                if ( find(v1vector.begin(), v1vector.end(), v2) == v1vector.end())
                    v1vector.push_back(v2);
            }
            lastPointIn = strokePoints[i];
        }
    }
    if (v0vsv1 > 0)
    {
        constraints0 = v0vector;
        constraints1 = v1vector;
    }
    else
    {
        constraints1 = v0vector;
        constraints0 = v1vector;
    }
    if (connectedCompRepresentatives.size() > 1 )
    {
        vector<bool> onComp;
        for (unsigned int i = 0; i < connectedCompRepresentatives.size(); i++)
            onComp.push_back(false);

        unsigned int v = vertexCloseToPoint(strokePoints[0], 0.4);
        onComp[vertexConnectedComp[v]] = true;
        for (unsigned int i = 0; i < onComp.size(); i++)
        {
            if (!onComp[i])
            {
                constraints1.push_back(connectedCompRepresentatives[i]);
                Trace (4, " add a constraints on "<< i << " component");
            }
        }
    }
}

void PartSelection::setHarmonicConstraints( unsigned int nb_of_ref_points,
                                            const std::vector<UserStroke::strokePoints> & strokePoints,
                                             std::vector<unsigned int> & constraints0,
                                             std::vector<unsigned int> & constraints1 )
{
    Trace(2, "PartSelection::setHarmonicConstraints");
    std::vector<Vec3Df> stroke3Dpoints;
    for (unsigned int i = 0; i < strokePoints.size(); i++)
        stroke3Dpoints.push_back(strokePoints[i].p);
    setHarmonicConstraints(nb_of_ref_points, stroke3Dpoints, constraints0, constraints1);
}


void PartSelection::initCholmod(bool constraints,
                                  std::vector<unsigned int> & constraints0,
                                  std::vector<unsigned int> & constraints1 )
{

    Trace(2, "PartSelection::initCholmod");
    // if there is no constraints given,
    // compute the harmonic field with random constraints
    // allow to factorize the harmonic field when the mesh is loaded
    // in order to go fast even for the first selection
    if ( ! constraints )
    {
        constraints0.clear();
        constraints1.clear();
        for (unsigned int i = 0; i < connectedCompRepresentatives.size(); i++)
        {
            constraints1.push_back(connectedCompRepresentatives[i]);
        }
        constraints0.push_back(vertices.size()/2);
    }

    cholmod_struct.freeSolution();
    cholmod_struct.free();

    cholmod_struct.start();
    // Matrice A = [ L WP]T, #lines = n + 2c, #columns = n
    cholmod_struct.setDimensions(vertices.size()+ constraints0.size() + constraints1.size(), vertices.size(), 1);

    //Number of nnz in L
    int _nb_non_zeros_in_A = vertices.size();
    for( unsigned int i = 0 ; i < oneRing.size(); i++ ){
        _nb_non_zeros_in_A += oneRing[i].size();
    }
    // + Number of nnz in WP
    _nb_non_zeros_in_A += constraints0.size() + constraints1.size();

    cholmod_struct.specifyNonZeroInA(_nb_non_zeros_in_A);
    cholmod_struct.allocate();

    //set the A matrix in cholmod
    // first, we enter all the values of the laplacian
    const vector< vector<pair <unsigned int, float> > > laplacianValues = _laplacian->getLaplacian();

    for( unsigned int i = 0; i < laplacianValues.size() ; i++ )
        for( unsigned int j = 0; j < laplacianValues[i].size(); j++ )
            cholmod_struct.addValueInA(i, laplacianValues[i][j].first, laplacianValues[i][j].second );

    // add constraints
    float constraint_w = CONSTRAINT_WEIGHT;

    // 0constraints goes from lines n to lines n+c
    for (unsigned int i = 0; i < constraints0.size(); i++)
        cholmod_struct.addValueInA(vertices.size() + i,  constraints0[i], constraint_w );

    // 1constraints goes from lines n+c to lines n+2c
    for (unsigned int i = 0; i < constraints1.size(); i++)
        cholmod_struct.addValueInA(vertices.size()+constraints0.size() + i,   constraints1[i], constraint_w );

    Trace(1, "--------before factorizing A in initCholmod");
    cholmod_struct.factorize();
    Trace(1, "--------after factorizing A in initCholmod");

    for (unsigned int i = 0; i < vertices.size(); i ++)
        cholmod_struct.setValueInB(i, 0, 0.);
    for (unsigned int i = 0; i < constraints0.size(); i++)
        cholmod_struct.setValueInB( vertices.size() + i, 0, constraint_w);

    for (unsigned int i = 0; i < constraints1.size(); i++)
         cholmod_struct.setValueInB( vertices.size() + constraints0.size() +i, 0, 0. );
    cholmod_struct.solve(false);
    cholmodInitialized = true;

    for (unsigned int i = 0; i < vertices.size(); i++)
        userDefinedHarmonicField[i] = cholmod_struct.getSolutionValue(i, 0);
}


// update = true : we need to update the factorization, and solve the system with the new b
// update = false : we only need to downdate the factorization
void PartSelection::setCholmodUpdate( bool update,
                                      std::vector<unsigned int> & _constraints0,
                                      std::vector<unsigned int> & _constraints1  )
{
    Trace(3, "PartSelection::setCholmodUpdate");
    cholmod_struct.freeUpdate();
    cholmod_struct.setCDimensions(_constraints0.size() + _constraints1.size(), vertices.size());
    cholmod_struct.specifyNonZeroInC(_constraints0.size() + _constraints1.size());
    cholmod_struct.allocateUpdate();

    // add constraints
    float constraint_w = CONSTRAINT_WEIGHT;

    for (unsigned int i = 0; i < _constraints0.size(); i++)
        cholmod_struct.addValueInC(i,  _constraints0[i], constraint_w );

    for (unsigned int i = 0; i <_constraints1.size(); i++)
        cholmod_struct.addValueInC(_constraints0.size() + i,   _constraints1[i], constraint_w );

    if (update)
    {
        for (unsigned int i = 0; i < _constraints0.size(); i++)
            cholmod_struct.setValueInDeltaB( i, 0,  constraint_w);

        for (unsigned int i = 0; i < _constraints1.size(); i++)
            cholmod_struct.setValueInDeltaB( _constraints0.size() +i, 0, 0. );
    }

    if (update)
    {
        cholmod_struct.freeSolution();
        cholmod_struct.update();
        for (unsigned int i = 0; i < vertices.size(); i++)
            userDefinedHarmonicField[i] = cholmod_struct.getSolutionValue(i,0);
    } else
        cholmod_struct.downdate();
}

void PartSelection::drawUserConstraints(){
    Vec3Df p;
    glColor3f(0.f, 1.f, 0.f);
    for (unsigned int i = 0; i < constraints0.size(); i++)
    {
        p = vertices[constraints0[i]].getPos();
        BasicGL::drawSphere(p[0], p[1], p[2], 0.005, 15, 15 );
    }
    glColor3f(1.f, 0.f, 0.f);
    for (unsigned int i = 0; i < constraints1.size(); i++)
    {
        p = vertices[constraints1[i]].getPos();
        BasicGL::drawSphere(p[0], p[1], p[2], 0.005, 15, 15 );
    }
}

void PartSelection::drawIsolines(){

    glDisable(GL_DEPTH_TEST);
    glEnable (GL_LINE_SMOOTH);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
    glLineWidth (1.5f);

    Vec3Df p;
    glColor3f(1.f, 0.f, 0.f);
    float r, g, b;
    for (unsigned int i = 0; i  < isolines.size(); i++)
    {
        RGB::calc_RGB(i ,0 , isolines.size() , r ,g , b );
        glColor3f(r,g,b);
        glBegin(GL_LINE_STRIP);
        for (unsigned int j = 0; j < isolines[i].size(); j++)
            glVertexVec3Df( isolines[i][j] );

        // to end the line
        if (!isolines[i].empty())
            glVertexVec3Df( isolines[i][0] );
        glEnd();
    }
    glDisable (GL_LINE_SMOOTH);
    glDisable (GL_BLEND);
    glEnable(GL_DEPTH_TEST);
}

void PartSelection::drawSimilarConstraints(){
    Vec3Df p;
    for (unsigned int i = 0; i < potentialSelections.size(); i++)
    {
        glColor3f(1-(float)i/(float)(potentialSelections.size()-1.f),
                  0.3*(float)i/(float)(potentialSelections.size()-1.f),
                  0.8*(float)i/(float)(potentialSelections.size()-1.f));

        for (unsigned int j = 0; j < potentialSelections[i].c0.size(); j++)
        {
            p = vertices[potentialSelections[i].c0[j]].getPos();
            BasicGL::drawSphere( p[0], p[1], p[2], 0.01,15,15 );
        }
        for (unsigned int j = 0; j < potentialSelections[i].c1.size(); j++)
        {
            p = vertices[potentialSelections[i].c1[j]].getPos();
            BasicGL::drawSphere( p[0], p[1], p[2], 0.005,15,15 );
        }
    }
}


void PartSelection::findSelectedFaces()
{
    Trace(3, "PartSelection::findSelectedFaces");
    // todo something smarter to find the best isoline
//    unsigned int refIsoline = ISOLINES_NB/2;
    double r;
    vector<bool> selected;

    for (unsigned int i = 0; i < triangles.size(); i++)
    {
        selected.push_back(false);
        faceInSelectionK[i] = -1.f;
        const Triangle &t = triangles[i];
        unsigned int j = 0;
        bool triangleSelected = false;
        while ( !triangleSelected && j < 3)
        {
            r = cholmod_struct.getSolutionValue(t.getVertex(j), 0);
            if  ( r >  (float) refIsoline/(float)(ISOLINES_NB+1) )
            {
                triangleSelected = true;
                selected[i] = true;
            }
            j++;
        }
    }

    std::vector<bool> inSelection;
    for (unsigned int i = 0; i < triangles.size(); i++)
        inSelection.push_back(false);

    if ( !isolines[refIsoline].empty())
    {
        Vec3Df iso_pnt ;
        unsigned int t;
        double d0, d1, d2;
        bool b = false;
        unsigned int i = 0;
        while (! b && i < isolines[refIsoline].size()) {
            iso_pnt = isolines[refIsoline][i];
            b = barycentricCoordinates(iso_pnt, t, d0, d1, d2);
            i++;
        }
        if (!b)
        {
            Trace(1, " ISSUE in findSelectedFaces");
            return;
        }
        inSelection[t] = true;

        queue<unsigned int> tToLook;
        tToLook.push(t);
        unsigned int cur_t;
        while (! tToLook.empty())
        {
            cur_t = tToLook.front();
            tToLook.pop();
            for (unsigned int i = 0; i < faceNeighbors[cur_t].size(); i++)
            {
                if ( ! inSelection[faceNeighbors[cur_t][i]] && selected[faceNeighbors[cur_t][i]])
                {
                    tToLook.push(faceNeighbors[cur_t][i]);
                    inSelection[faceNeighbors[cur_t][i]] = true;
                }
            }
        }
    }
    for (unsigned int i = 0; i < inSelection.size(); i++)
        if (inSelection[i])
            faceInSelectionK[i] = 1000.f;
}



QPixmap PartSelection::getUnderMouseHistogram(unsigned int v, float &dist){
    QPixmap ret;
    int f = triangleCloseToPoint(vertices[v].getPos(), 0.2);
    if (f >= 0 && f < (int) triangles.size())
    {
        for (unsigned int i = 0; i < potentialSelections.size(); i++)
        {
                if (find(potentialSelections[i].faces.begin(), potentialSelections[i].faces.end(),f)
                        != potentialSelections[i].faces.end())
                {
                    dist = potentialSelections[i].dist_to_ref;
                    return potentialSelections[i].hist.drawHistogram();
                }
        }
    }
    return ret;
}

